import os
import platform
import shutil
import sys
import unittest
from os.path import abspath, dirname, join
from subprocess import Popen
from tempfile import NamedTemporaryFile
from time import sleep
from unittest.case import skipIf

import requests
from requests.exceptions import ConnectionError
from requests.models import Response

import eventsourcing.example.interface
from eventsourcing.domain.model.events import assert_event_handlers_empty
from eventsourcing.example.application import close_example_application
from eventsourcing.infrastructure.sqlalchemy.datastore import (
    SQLAlchemyDatastore,
    SQLAlchemySettings,
)
from eventsourcing.tests.base import notquick

path_to_virtualenv = None
if hasattr(sys, "real_prefix"):
    path_to_virtualenv = sys.prefix

path_to_eventsourcing = dirname(dirname(abspath(eventsourcing.__file__)))
path_to_interface_module = dirname(abspath(eventsourcing.example.interface.__file__))
path_to_flaskapp = join(path_to_interface_module, "flaskapp.py")


@notquick
class TestFlaskApp(unittest.TestCase):
    port = 5001

    def setUp(self):
        assert_event_handlers_empty()
        super(TestFlaskApp, self).setUp()
        self.app = self.start_app()

    def start_app(self):
        cmd = [sys.executable, path_to_flaskapp]
        return Popen(cmd)

    def tearDown(self):
        self.app.terminate()
        sleep(1)
        self.app.kill()
        sleep(1)
        self.app.wait()
        super(TestFlaskApp, self).tearDown()
        assert_event_handlers_empty()

    def test(self):
        patience = 100
        while True:
            try:
                response = requests.get("http://localhost:{}".format(self.port))
                break
            except ConnectionError:
                patience -= 1
                if not patience:
                    self.fail(
                        "Couldn't get response from app, (Python executable {})".format(
                            sys.executable
                        )
                    )
                else:
                    sleep(0.1)

        self.assertIsInstance(response, Response)
        self.assertIn("Hello There!", response.text)


@notquick
@skipIf(
    platform.python_implementation() == "PyPy",
    "uWSGI needs special plugin to run with PyPy",
)
class TestFlaskWsgi(TestFlaskApp):
    port = 9001

    def start_app(self):
        # Make up a DB URI using a named temporary file.
        self.tempfile = NamedTemporaryFile()
        uri = "sqlite:///{}".format(self.tempfile.name)

        from eventsourcing.example.interface.flaskapp import IntegerSequencedItem

        # Close application, importing the module constructed
        # the application, which will leave event handlers subscribed.
        close_example_application()

        # Setup tables.
        datastore = SQLAlchemyDatastore(
            settings=SQLAlchemySettings(uri=uri), tables=(IntegerSequencedItem,)
        )
        datastore.setup_connection()
        datastore.setup_tables()
        datastore.close_connection()

        # Run uwsgi.
        path_to_uwsgi = shutil.which("uwsgi")
        if not os.path.exists(path_to_uwsgi):
            raise AssertionError("Can't find uwsgi: %s" % path_to_uwsgi)
        cmd = [path_to_uwsgi]
        if path_to_virtualenv is not None:
            cmd += ["-H", path_to_virtualenv]
        cmd += ["--master"]
        cmd += ["--processes", "4"]
        cmd += ["--threads", "2"]
        cmd += ["--wsgi-file", path_to_flaskapp]
        cmd += ["--http", ":{}".format(self.port)]
        pythonpath = ":".join(
            os.getenv("PYTHONPATH", "").split(":") + [path_to_eventsourcing]
        )
        return Popen(cmd, env={"PYTHONPATH": pythonpath, "DB_URI": uri})



import time
from abc import abstractmethod
from collections import defaultdict, deque
from queue import Empty, Queue
from threading import Barrier, BrokenBarrierError, Event, Lock, Thread, Timer
from time import sleep
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Deque,
    Dict,
    Generic,
    List,
    Optional,
    Type,
    Union,
    no_type_check,
)

from eventsourcing.application.notificationlog import NotificationLogReader
from eventsourcing.application.process import ProcessApplication, PromptToQuit
from eventsourcing.application.simple import (
    ApplicationWithConcreteInfrastructure,
    Prompt,
    PromptToPull,
    is_prompt_to_pull,
)
from eventsourcing.domain.model.decorators import retry
from eventsourcing.domain.model.events import subscribe, unsubscribe
from eventsourcing.exceptions import (
    CausalDependencyFailed,
    EventSourcingError,
    OperationalError,
    ProgrammingError,
    RecordConflictError,
)
from eventsourcing.system.definition import AbstractSystemRunner, System
from eventsourcing.whitehead import T

DEFAULT_POLL_INTERVAL = 5


# Todo: Support passing SQLAlchemy session into runner.
class InProcessRunner(AbstractSystemRunner):
    """
    Runs a system in the current process,
    either in the current thread, or with
    one thread for each process in the system.
    """

    def start(self) -> None:
        if len(self.processes):
            raise ProgrammingError("Already running")

        # Construct the processes.
        for process_class in self.system.process_classes.values():
            self._construct_app_by_class(process_class)

        # Tell each process which other processes to follow.
        for downstream_name, upstream_names in self.system.upstream_names.items():
            downstream_process = self.processes[downstream_name]
            for upstream_name in upstream_names:
                upstream_process = self.processes[upstream_name]
                upstream_log = upstream_process.notification_log
                downstream_process.follow(upstream_name, upstream_log)

        # Do something to propagate prompts.
        subscribe(predicate=is_prompt_to_pull, handler=self.handle_prompt)

    @abstractmethod
    def handle_prompt(self, prompt: Prompt) -> None:
        """
        Handles publication of a prompt.

        Abstract method which must be implemented on concrete descendants.
        """

    def close(self) -> None:
        super(InProcessRunner, self).close()

        unsubscribe(predicate=is_prompt_to_pull, handler=self.handle_prompt)


class SingleThreadedRunner(InProcessRunner):
    """
    Runs a system in the current thread.
    """

    def __init__(
        self,
        system: System,
        infrastructure_class: Optional[
            Type[ApplicationWithConcreteInfrastructure]
        ] = None,
        **kwargs: Any,
    ):
        super(SingleThreadedRunner, self).__init__(
            system, infrastructure_class=infrastructure_class, **kwargs
        )
        if TYPE_CHECKING:
            self.pending_prompts: Queue[Prompt]
        self.pending_prompts = Queue()

        self.iteration_lock = Lock()

    def handle_prompt(self, prompt: Prompt) -> None:
        self.run_followers(prompt)

    def run_followers(self, prompt: Prompt) -> None:
        """
        First caller adds a prompt to queue and
        runs followers until there are no more
        pending prompts.

        Subsequent callers just add a prompt
        to the queue, avoiding recursion.
        """
        # Put the prompt on the queue.
        self.pending_prompts.put(prompt)

        if self.iteration_lock.acquire(False):
            # start_time = time.time()
            i = 0
            try:
                while True:
                    try:
                        prompt = self.pending_prompts.get(False)
                    except Empty:
                        break
                    else:
                        if isinstance(prompt, PromptToPull):
                            downstream_names = self.system.downstream_names[
                                prompt.process_name
                            ]
                            for downstream_name in downstream_names:
                                downstream_process = self.processes[downstream_name]
                                downstream_process.run(prompt)
                                i += 1
                        self.pending_prompts.task_done()
            finally:
                # run_frequency = i / (time.time() - start_time)
                # print(f"Run frequency: {run_frequency}")
                self.iteration_lock.release()


class MultiThreadedRunner(InProcessRunner):
    """
    Runs a system with a thread for each process.
    """

    def __init__(
        self,
        system: System,
        poll_interval: Optional[int] = None,
        clock_speed: Optional[Union[int, float]] = None,
        **kwargs: Any,
    ):
        super(MultiThreadedRunner, self).__init__(system=system, **kwargs)
        self.poll_interval = poll_interval or DEFAULT_POLL_INTERVAL
        assert isinstance(system, System)
        self.threads: Dict[str, PromptQueuedApplicationThread] = {}
        self.clock_speed: Optional[Union[float, int]] = clock_speed
        if self.clock_speed:
            self.clock_event = Event()
            self.stop_clock_event = Event()

        self.inboxes: Dict[str, Queue] = {}
        self.outboxes: Dict[str, PromptOutbox[str]] = {}

    def start(self) -> None:
        super(MultiThreadedRunner, self).start()
        assert not self.threads, "Already started"
        assert not self.inboxes, "Already has inboxes"
        assert not self.outboxes, "Already has outboxes"

        # Setup queues.
        for process_name, upstream_names in self.system.upstream_names.items():
            inbox_id = process_name.lower()
            if inbox_id not in self.inboxes:
                self.inboxes[inbox_id] = Queue()
            for upstream_class_name in upstream_names:
                outbox_id = upstream_class_name.lower()
                if outbox_id not in self.outboxes:
                    self.outboxes[outbox_id] = PromptOutbox()
                if inbox_id not in self.outboxes[outbox_id].downstream_inboxes:
                    self.outboxes[outbox_id].downstream_inboxes[
                        inbox_id
                    ] = self.inboxes[inbox_id]

        # Construct application threads.
        for process_name, process in self.processes.items():
            if self.clock_speed:
                process.clock_event = self.clock_event
                process.tick_interval = 1 / self.clock_speed

            thread = PromptQueuedApplicationThread(
                process=process,
                poll_interval=self.poll_interval,
                inbox=self.inboxes[process_name],
                outbox=self.outboxes.get(process_name),
                # Todo: Is it better to clock the prompts or the notifications?
                # clock_event=clock_event
            )
            self.threads[process_name] = thread

        # Start application threads.
        for thread in self.threads.values():
            thread.start()

        for thread in self.threads.values():
            thread.is_running.wait()

        # Start clock.
        if self.clock_speed:
            self.start_clock()

    def start_clock(self) -> None:
        assert self.clock_speed, self.clock_speed
        tick_interval = 1 / self.clock_speed
        # print(f"Tick interval: {tick_interval:.6f}s")
        self.last_tick = None
        self.this_tick = None
        self.tick_adjustment = 0

        def set_clock_event() -> None:
            if self.stop_clock_event.is_set():
                return

            self.this_tick = time.process_time()

            if self.last_tick:
                tick_size = self.this_tick - self.last_tick

                tick_oversize = tick_size - tick_interval
                tick_oversize_percentage = 100 * (tick_oversize) / tick_interval
                if tick_oversize_percentage > 300:
                    print(
                        f"Warning: Tick over size: {tick_size :.6f}s "
                        f"{tick_oversize_percentage:.2f}%"
                    )

                if abs(tick_oversize_percentage) < 300:
                    self.tick_adjustment += 0.5 * tick_interval * tick_oversize
                    max_tick_adjustment = 0.5 * tick_interval
                    min_tick_adjustment = 0
                    self.tick_adjustment = min(
                        self.tick_adjustment, max_tick_adjustment
                    )
                    self.tick_adjustment = max(
                        self.tick_adjustment, min_tick_adjustment
                    )

            self.last_tick = self.this_tick

            self.clock_event.set()
            self.clock_event.clear()

            if not self.stop_clock_event.is_set():
                set_timer()

        def set_timer() -> None:
            # print(f"Tick adjustment: {self.tick_adjustment:.6f}")
            if self.last_tick is not None:
                time_since_last_tick = time.process_time() - self.last_tick
                time_remaining = tick_interval - time_since_last_tick
                timer_interval = time_remaining - self.tick_adjustment
                if timer_interval < 0:
                    timer_interval = 0
                    # print("Warning: clock thread is running flat out!")
            else:
                timer_interval = 0
            timer = Timer(timer_interval, set_clock_event)
            timer.start()

        set_timer()

    def handle_prompt(self, prompt: Prompt) -> None:
        self.broadcast_prompt(prompt)

    def broadcast_prompt(self, prompt: Prompt) -> None:
        if isinstance(prompt, PromptToPull):
            outbox_id = prompt.process_name
            outbox = self.outboxes.get(outbox_id)
            if outbox:
                outbox.put(prompt)

    def close(self) -> None:
        super(MultiThreadedRunner, self).close()

        if self.clock_speed:
            self.clock_event.set()
            self.stop_clock_event.set()

        for thread in self.threads.values():
            thread.inbox.put(PromptToQuit())

        for thread in self.threads.values():
            thread.join(timeout=10)

        self.threads.clear()
        self.outboxes.clear()
        self.inboxes.clear()


class PromptOutbox(Generic[T]):
    """
    Has a collection of downstream prompt inboxes.

    """

    def __init__(self) -> None:
        self.downstream_inboxes: Dict[T, Queue] = {}

    def put(self, prompt: Union[PromptToPull, str]) -> None:
        """
        Puts prompt in each downstream inbox (an actual queue).
        """
        for queue in self.downstream_inboxes.values():
            queue.put(prompt)


class PromptQueuedApplicationThread(Thread):
    """
    Application thread which uses queues of prompts.

    It loops on an "inbox" queue of prompts, and
    adds its prompts to an "outbox" queue.
    """

    def __init__(
        self,
        *,
        process: ProcessApplication,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
        inbox: Queue,
        outbox: Optional[PromptOutbox],
        clock_event: Optional[Event] = None,
    ):
        """
        Initialises the thread with a process application object, and
        inbox and outbox.

        :param ProcessApplication process: Application object.
        :param int poll_interval: Interval to check for upstream events.
        :param Queue inbox: For incoming prompts.
        :param PromptOutbox outbox: For outgoing prompts.
        :param clock_event: Event that "clocks" this thread (optional).
        """
        super(PromptQueuedApplicationThread, self).__init__(daemon=True)
        self.app = process
        self.poll_interval = poll_interval
        if TYPE_CHECKING:
            self.inbox: Queue[Union[Prompt, str]]
        self.inbox = inbox
        self.outbox = outbox
        self.clock_event = clock_event
        self.is_running = Event()

    def run(self) -> None:
        self.loop_on_prompts()

    def loop_on_prompts(self) -> None:

        # Loop on getting prompts.
        self.is_running.set()
        while True:
            try:
                # Todo: Make the poll interval gradually
                #  increase if there are only timeouts?
                prompt = self.inbox.get(timeout=self.poll_interval)

            except Empty:
                # Basically, we're polling after a timeout.
                if self.clock_event is None:
                    self.run_process()

            else:
                self.inbox.task_done()

                if isinstance(prompt, PromptToQuit):
                    self.app.close()
                    break

                elif isinstance(prompt, PromptToPull):
                    started = None
                    if self.clock_event is not None:
                        self.clock_event.wait()
                        started = time.time()

                    self.run_process(prompt)

                    if self.clock_event is not None:
                        assert started
                        ended = time.time()
                        duration = ended - started
                        if self.clock_event.is_set():
                            print(
                                f"Warning: Process {self.app.name} overran clock "
                                f"cycle: {duration}"
                            )
                        else:
                            print(
                                f"Info: Process {self.app.name} ran within clock "
                                f"cycle: {duration}"
                            )
                else:
                    raise Exception("Unsupported prompt: {}".format(prompt))

    def run_process(self, prompt: Optional[PromptToPull] = None) -> None:
        try:
            self._run_process(prompt)
        except CausalDependencyFailed:
            pass
        except EventSourcingError:
            pass

    @retry(CausalDependencyFailed, max_attempts=100, wait=0.2)
    @retry((OperationalError, RecordConflictError), max_attempts=100, wait=0.01)
    def _run_process(self, prompt: Optional[PromptToPull] = None) -> None:
        self.app.run(prompt)


class SteppingRunner(InProcessRunner):
    def __init__(
        self,
        normal_speed: int = 1,
        scale_factor: int = 1,
        is_verbose: bool = False,
        *args: Any,
        **kwargs: Any,
    ):
        super(SteppingRunner, self).__init__(*args, **kwargs)
        self.normal_speed = normal_speed
        self.scale_factor = scale_factor
        self.is_verbose = is_verbose
        if scale_factor:
            self.tick_interval = 1 / (normal_speed * scale_factor)
        else:
            self.tick_interval = 0
        if self.is_verbose:
            print(f"Tick interval: {self.tick_interval:.6f}s")
        self.clock_thread: Optional[ClockThread] = None

    def call_in_future(self, cmd: Callable, ticks_delay: int) -> None:
        assert self.clock_thread
        self.clock_thread.call_in_future(cmd, ticks_delay)


class SteppingSingleThreadedRunner(SteppingRunner):
    def __init__(self, *args: Any, **kwargs: Any):
        super(SteppingSingleThreadedRunner, self).__init__(*args, **kwargs)
        self.seen_prompt_events: Dict[str, Event] = {}
        self.stop_event = Event()

    def start(self) -> None:
        super(SteppingSingleThreadedRunner, self).start()

        self.clock_thread = ProcessRunningClockThread(
            normal_speed=self.normal_speed,
            scale_factor=self.scale_factor,
            stop_event=self.stop_event,
            is_verbose=self.is_verbose,
            seen_prompt_events=self.seen_prompt_events,
            processes=self.processes,
            use_direct_query_if_available=self.use_direct_query_if_available,
        )
        self.clock_thread.start()

    def handle_prompt(self, prompt: Prompt) -> None:
        """
        Ignores prompts.
        """

    def close(self) -> None:
        self.stop_event.set()
        super(SteppingSingleThreadedRunner, self).close()
        if self.clock_thread:
            self.clock_thread.join(5)
            if self.clock_thread.is_alive():
                print("Warning: clock thread was still alive")


class ClockThread(Thread):
    def __init__(self, *args: Any, **kwargs: Any):
        super(ClockThread, self).__init__()
        self.future_cmds: Dict[int, List[Callable]] = defaultdict(list)
        self.tick_count = 0

    def call_in_future(self, cmd: Callable, ticks_delay: int) -> None:
        ticks_delay = max(ticks_delay, 1)
        self.future_cmds[ticks_delay + self.tick_count].append(cmd)

    def call_commands(self) -> None:
        for future_cmd in self.future_cmds.get(self.tick_count, []):
            future_cmd()

    @no_type_check
    def do_tick(self, process_time: float, tick_time: float) -> None:
        # Todo: Pull these members up from subclasses.
        tick_duration = tick_time - self.last_tick_time
        self.all_tick_durations.append(tick_duration)
        if len(self.all_tick_durations) > self.tick_durations_window_size:
            self.all_tick_durations.popleft()
        if self.is_verbose:
            process_duration = process_time - self.last_process_time
            intensity = 100 * process_duration / tick_duration
            clock_speed = 1 / tick_duration
            real_time = self.tick_count / self.normal_speed
            print(
                f"Tick {self.tick_count:4}: {real_time:4.2f}s  "
                f"{tick_duration:.6f}s, "
                f"{intensity:6.2f}%, {clock_speed:6.1f}Hz, "
                f"{self.actual_clock_speed:6.1f}Hz, "
                f"{self.tick_adjustment:.6f}s"
            )
        if self.tick_interval:
            tick_oversize = tick_duration - self.tick_interval
            tick_oversize_percentage = 100 * (tick_oversize) / self.tick_interval
            # if tick_oversize_percentage > 300:
            #     print(f"Warning: Tick over size: { tick_duration :.6f}s
            #     {tick_oversize_percentage:.2f}%")

            if abs(tick_oversize_percentage) < 300:
                # Weight falls from 1 as reciprocal of count, to tick
                # interval.
                # weight = max(1 / self.tick_count, min(.1,
                # self.tick_interval))
                weight = 1 / (1 + self.tick_count * self.tick_interval) ** 2
                # print(f"Weight: {weight:.4f}")
                self.tick_adjustment += weight * tick_oversize
                max_tick_adjustment = 1.0 * self.tick_interval
                min_tick_adjustment = 0
                self.tick_adjustment = min(self.tick_adjustment, max_tick_adjustment)
                self.tick_adjustment = max(self.tick_adjustment, min_tick_adjustment)


class ProcessRunningClockThread(ClockThread):
    def __init__(
        self,
        *,
        normal_speed: int,
        scale_factor: int,
        stop_event: Event,
        is_verbose: bool = False,
        seen_prompt_events: Dict[str, Event],
        processes: Dict[str, ProcessApplication],
        use_direct_query_if_available: bool = False,
    ):
        super(ProcessRunningClockThread, self).__init__(daemon=True)
        self.normal_speed = normal_speed
        self.scale_factor = scale_factor
        self.stop_event = stop_event
        self.seen_prompt_events = seen_prompt_events
        self.processes = processes
        self.use_direct_query_if_available = use_direct_query_if_available
        self.last_tick_time: Union[float, int] = 0
        self.last_process_time = 0.0
        self.all_tick_durations: Deque = deque()
        self.tick_adjustment: float = 0.0
        self.is_verbose = is_verbose
        if normal_speed and scale_factor:
            self.tick_interval = 1 / normal_speed / scale_factor
            self.tick_durations_window_size = max(
                100, int(round(1 / self.tick_interval, 0))
            )
        else:
            self.tick_interval = 0
            self.tick_durations_window_size = 1000
        # Construct lists of followers for each process.
        self.followers: Dict[str, List[str]] = {}
        for process in self.processes.values():
            self.followers[process.name] = []
        for process in self.processes.values():
            for upstream_process_name in process.readers:
                self.followers[upstream_process_name].append(process.name)

        # Construct a notification log reader for each process.
        self.readers: Dict[str, NotificationLogReader] = {}
        for process in self.processes.values():
            reader = process.notification_log_reader_class(
                notification_log=process.notification_log,
                use_direct_query_if_available=self.use_direct_query_if_available,
            )
            self.readers[process.name] = reader

    @property
    def actual_clock_speed(self) -> Union[int, float]:
        if self.all_tick_durations:
            # Todo: Might need to lock this, to avoid "RuntimeError: deque mutated
            #  during iteration".
            durations = list(self.all_tick_durations)
            return len(durations) / sum(durations)
        else:
            return 0

    def run(self) -> None:
        # Get new notifications once.

        # loop_counter = 0

        while not self.stop_event.is_set():
            # print("Loop count: {}".format(loop_counter))
            try:
                # Get all notifications.
                all_notifications = {}
                for process_name in self.processes:
                    # seen_prompt = self.seen_prompt_events[process_name]
                    # if seen_prompt.is_set():
                    #     seen_prompt.clear()
                    reader = self.readers[process_name]
                    notifications = reader.list_notifications()
                    all_notifications[process_name] = notifications

                # Process all notifications.
                all_events = {}
                for process_name, notifications in all_notifications.items():
                    events = []
                    for notification in notifications:
                        # print(process_name, notification)
                        process = self.processes[process_name]
                        # It's not the follower process, but the method does the same
                        # thing.
                        event = process.get_event_from_notification(notification)
                        notification_id = notification["id"]
                        events.append((notification_id, event))
                    all_events[process_name] = events

                for process_name, events in all_events.items():
                    # print(f"Process: {process_name}")
                    for follower_name in self.followers[process_name]:
                        follower_process = self.processes[follower_name]
                        # print(f"Follower: {follower_name}")
                        for notification_id, event in events:
                            # print(f"Notification: {notification_id}, {event}")
                            follower_process.process_upstream_event(
                                event, notification_id, process_name
                            )
                # Call commands delayed until this clock tick.
                self.call_commands()

            except:
                if self.stop_event.is_set():
                    break
                else:
                    # print("Error... Loop count: {}".format(loop_counter))
                    self.stop_event.set()
                    raise
            else:
                tick_time: float = time.time()
                process_time = time.process_time()
                if self.last_tick_time:
                    self.do_tick(process_time, tick_time)

                self.last_tick_time = tick_time
                self.last_process_time = process_time
                self.tick_count += 1

                if self.tick_interval:
                    sleep_interval = self.tick_interval - self.tick_adjustment
                    sleep(max(sleep_interval, 0))


class SteppingMultiThreadedRunner(SteppingRunner):
    """
    Has a clock thread, and a thread for each application process
    in the system. The clock thread loops until stopped, waiting
    for a barrier, after sleeping for remaining tick interval timer.
    Application threads loop until stopped, waiting for the same
    barrier. Then, after all threads are waiting at the barrier,
    the barrier is lifted. The clock thread proceeds by sleeping
    for the clock tick interval. The application threads proceed by
    getting new notifications and processing all of them.

    There are actually two barriers, so that each application thread
    waits before getting notifications, and then waits for all processes
    to complete getting notification before processing the notifications
    through the application policy. This avoids events created by a process
    application "bleeding" into the notifications of another process
    application in the same clock cycle.


    Todo:
    Receive prompts, but set an event for the prompting process, to avoid unnecessary
    runs.

    Allow commands to be scheduled at future clock tick number, and execute when
    reached.

    """

    def __init__(self, *args: Any, **kwargs: Any):
        super(SteppingMultiThreadedRunner, self).__init__(*args, **kwargs)
        self.seen_prompt_events: Dict[str, Event] = {}
        self.fetch_barrier: Optional[Barrier] = None
        self.execute_barrier: Optional[Barrier] = None
        self.application_threads: Dict[str, BarrierControlledApplicationThread] = {}
        self.clock_thread = None
        self.stop_event = Event()

    def handle_prompt(self, prompt: Prompt) -> None:
        if isinstance(prompt, PromptToPull):
            seen_prompt = self.seen_prompt_events[prompt.process_name]
            seen_prompt.set()

    def start(self) -> None:
        super(SteppingMultiThreadedRunner, self).start()
        parties = 1 + len(self.processes)
        self.fetch_barrier = Barrier(parties)
        self.execute_barrier = Barrier(parties)

        # Create an event for each process.
        for process_name in self.processes:
            self.seen_prompt_events[process_name] = Event()

        # Construct application threads.
        for process_name, process in self.processes.items():
            process_instance_id = process_name

            thread = BarrierControlledApplicationThread(
                process=process,
                fetch_barrier=self.fetch_barrier,
                execute_barrier=self.execute_barrier,
                stop_event=self.stop_event,
            )
            self.application_threads[process_instance_id] = thread

        # Start application threads.
        for thread in self.application_threads.values():
            thread.start()

        # Start clock thread.
        self.clock_thread = BarrierControllingClockThread(
            normal_speed=self.normal_speed,
            scale_factor=self.scale_factor,
            tick_interval=self.tick_interval,
            fetch_barrier=self.fetch_barrier,
            execute_barrier=self.execute_barrier,
            stop_event=self.stop_event,
            is_verbose=self.is_verbose,
        )
        self.clock_thread.start()

    def close(self) -> None:
        self.stop_event.set()
        super(SteppingMultiThreadedRunner, self).close()
        if self.execute_barrier:
            self.execute_barrier.abort()
        if self.fetch_barrier:
            self.fetch_barrier.abort()

        for thread in self.application_threads.values():
            thread.join(timeout=1)
            if thread.is_alive():
                print(
                    f"Warning: application thread '"
                    f"{thread.app.name}"
                    f"' was still alive."
                )

        self.application_threads.clear()

        if self.clock_thread:
            self.clock_thread.join(timeout=1)
            if self.clock_thread.is_alive():
                print(f"Warning: clock thread was still alive")


class BarrierControlledApplicationThread(Thread):
    def __init__(
        self,
        process: ProcessApplication,
        fetch_barrier: Barrier,
        execute_barrier: Barrier,
        stop_event: Event,
    ):
        super(BarrierControlledApplicationThread, self).__init__(daemon=True)
        self.app = process
        self.fetch_barrier = fetch_barrier
        self.execute_barrier = execute_barrier
        self.stop_event = stop_event

    def run(self) -> None:
        while not self.stop_event.is_set():
            # Isolate "fetch" and "execute" steps, to avoid
            # events created in one application being processed by
            # another application in the same tick. Race condition
            # where one process writes new events before another has
            # read all notifications from last tick. Actually, need
            # to get all notifications from all upstream applications
            # and then process the notifications. The run() method
            # gets notifications from one, then processes, then gets
            # from another, which makes the race condition probable.
            all_notifications = []
            try:
                self.fetch_barrier.wait()
            except BrokenBarrierError:
                self.abort()
            else:
                try:
                    # Get all notifications.
                    for upstream_name in self.app.readers:
                        notifications = list(self.app.read_reader(upstream_name))
                        all_notifications.append((upstream_name, notifications))

                except:
                    self.abort()
                    raise

            if self.stop_event.is_set():
                break

            try:
                self.execute_barrier.wait()
            except BrokenBarrierError:
                self.abort()
            else:
                try:
                    # Process all notifications.
                    for upstream_name, notifications in all_notifications:
                        for notification in notifications:
                            event = self.app.get_event_from_notification(notification)
                            self.app.process_upstream_event(
                                event, notification["id"], upstream_name
                            )
                except:
                    self.abort()
                    raise

            try:
                self.execute_barrier.wait()
            except BrokenBarrierError:
                self.abort()

    def abort(self) -> None:
        self.stop_event.set()
        self.fetch_barrier.abort()
        self.execute_barrier.abort()


class BarrierControllingClockThread(ClockThread):
    def __init__(
        self,
        normal_speed: int,
        scale_factor: int,
        tick_interval: Optional[Union[int, float]],
        fetch_barrier: Barrier,
        execute_barrier: Barrier,
        stop_event: Event,
        is_verbose: bool = False,
    ):
        super(BarrierControllingClockThread, self).__init__(daemon=True)
        # Todo: Remove the redundancy here.
        self.normal_speed = normal_speed
        self.scale_factor = scale_factor
        self.tick_interval = tick_interval
        self.fetch_barrier = fetch_barrier
        self.execute_barrier = execute_barrier
        self.stop_event = stop_event
        self.last_tick_time = 0.0
        self.last_process_time = 0.0
        self.all_tick_durations: Deque = deque()
        self.tick_adjustment: float = 0.0
        self.is_verbose = is_verbose
        if self.tick_interval:

            self.tick_durations_window_size = max(
                1, int(round(1 / self.tick_interval, 0))
            )
        else:
            self.tick_durations_window_size = 100

    @property
    def actual_clock_speed(self) -> float:
        if self.all_tick_durations:
            durations = self.all_tick_durations
            return len(durations) / sum(durations)
        else:
            return 0.0

    def run(self) -> None:
        while not self.stop_event.is_set():

            try:
                self.fetch_barrier.wait()
                self.execute_barrier.wait()
                self.execute_barrier.wait()
                self.call_commands()
            except BrokenBarrierError:
                self.fetch_barrier.abort()
                self.execute_barrier.abort()
                self.stop_event.set()
            else:
                tick_time = time.time()
                process_time = time.process_time()
                if self.last_tick_time:
                    self.do_tick(process_time, tick_time)

                self.last_tick_time = tick_time
                self.last_process_time = process_time
                self.tick_count += 1

                if self.tick_interval:
                    sleep_interval = self.tick_interval - self.tick_adjustment
                    sleep(max(sleep_interval, 0))



from decimal import Decimal
from typing import Dict
from uuid import UUID, uuid1, uuid4

from eventsourcing.domain.model.events import (
    DomainEvent,
    EventHandlersNotEmptyError,
    EventWithHash,
    EventWithOriginatorID,
    EventWithOriginatorVersion,
    EventWithTimestamp,
    EventWithTimeuuid,
    assert_event_handlers_empty,
    clear_event_handlers,
    create_timesequenced_event_id,
    publish,
    subscribe,
    unsubscribe,
)
from eventsourcing.example.domainmodel import Example
from eventsourcing.exceptions import EventHashError, TopicResolutionError
from eventsourcing.utils.times import decimaltimestamp, decimaltimestamp_from_uuid
from eventsourcing.utils.topic import get_topic, resolve_topic

try:
    from unittest import mock, TestCase
except ImportError:
    import mock


class Event(DomainEvent):
    pass


# Check not equal to different type with same values.
class SubclassEvent(Event):
    pass


class TestAbstractDomainEvent(TestCase):
    def test(self):
        # Check base class can be sub-classed.

        # Check subclass can be instantiated.
        event1 = Event()
        self.assertEqual(type(event1).__qualname__, "Event")

        # Check subclass can be instantiated with other parameters.
        event2 = Event(name="value")

        # Check the attribute value is available.
        self.assertEqual(event2.name, "value")

        # Check the attribute value cannot be changed
        with self.assertRaises(AttributeError):
            event2.name = "another value"
        self.assertEqual(event2.name, "value")

        # Check it's equal to itself, by value.
        self.assertEqual(event2, event2)
        self.assertEqual(event2, Event(name="value"))

        # Check not equal to same event type with different values.
        self.assertNotEqual(event2, Event(name="another value"))

        self.assertNotEqual(event2, SubclassEvent(name=event2.name))


class TestEventWithOriginatorID(TestCase):
    def test(self):
        # Check base class can be sub-classed.
        class Event(EventWithOriginatorID):
            pass

        # Check can't instantiate without an ID.
        with self.assertRaises(TypeError):
            # noinspection PyArgumentList
            Event()

        # Check any kind of ID is acceptable.
        event = Event(originator_id="1")
        self.assertEqual(event.originator_id, "1")

        event = Event(originator_id=1)
        self.assertEqual(event.originator_id, 1)

        event = Event(originator_id=uuid4())
        self.assertIsInstance(event.originator_id, UUID)

        # Check the ID value can't be reassigned.
        with self.assertRaises(AttributeError):
            # noinspection PyPropertyAccess
            event.originator_id = "2"


class TestEventWithOriginatorVersion(TestCase):
    def test(self):
        # Check base class can be sub-classed.
        class Event(EventWithOriginatorVersion):
            pass

        # Check event can be instantiated with a version.
        version1 = 1
        event = Event(originator_version=version1)
        self.assertEqual(event.originator_version, version1)

        # Check event can't be instantiated without a version.
        with self.assertRaises(TypeError):
            # noinspection PyArgumentList
            event = Event()

        # Check version must be an integer.
        with self.assertRaises(TypeError):
            event = Event(originator_version="1")

        # Check the version value can't be reassigned.
        with self.assertRaises(AttributeError):
            # noinspection PyPropertyAccess
            event.originator_version = 2


class TestEventWithTimestamp(TestCase):
    def test(self):
        # Check base class can be sub-classed.
        class Event(EventWithTimestamp):
            pass

        # Check event can be instantiated with a timestamp.
        time1 = decimaltimestamp()
        event = Event(timestamp=time1)
        self.assertEqual(event.timestamp, time1)

        # Check event can be instantiated without a timestamp.
        event = Event()
        self.assertGreaterEqual(event.timestamp, time1)
        self.assertLessEqual(event.timestamp, decimaltimestamp())

        # Check the timestamp value can't be reassigned.
        with self.assertRaises(AttributeError):
            # noinspection PyPropertyAccess
            event.timestamp = decimaltimestamp()


class TestEventWithTimeuuid(TestCase):
    def test(self):
        # Check base class can be sub-classed.
        class Event(EventWithTimeuuid):
            pass

        # Check event has a UUID event_id.
        event = Event()
        self.assertIsInstance(event.event_id, UUID)

        # Check the event_id can't be reassigned.
        with self.assertRaises(AttributeError):
            # noinspection PyPropertyAccess
            event.event_id = decimaltimestamp()

        # Check event can be instantiated with a given UUID.
        event_id = uuid1()
        event = Event(event_id=event_id)
        self.assertEqual(event.event_id, event_id)

        # Generate a series of timestamps.
        events = [Event() for _ in range(100)]
        timestamps = [decimaltimestamp_from_uuid(e.event_id) for e in events]

        # Check series doesn't decrease at any point.
        last = timestamps[0]
        for timestamp in timestamps[1:]:
            self.assertLessEqual(last, timestamp)
            last = timestamp

        # Check last timestamp is greater than the first.
        self.assertGreater(timestamps[-1], timestamps[0])


class TestEventWithOriginatorVersionAndID(TestCase):
    # noinspection PyArgumentList
    def test(self):
        # Check base class can be sub-classed.
        class Event(EventWithOriginatorVersion, EventWithOriginatorID):
            pass

        # Check construction requires both an ID and version.
        with self.assertRaises(TypeError):
            # noinspection PyArgumentList
            Event()

        with self.assertRaises(TypeError):
            # noinspection PyArgumentList
            Event(originator_id="1")

        with self.assertRaises(TypeError):
            Event(originator_version=1)

        event1 = Event(originator_id="1", originator_version=0)

        event2 = Event(originator_id="1", originator_version=1)

        # Check the event attributes.
        self.assertEqual(event1.originator_id, "1")
        self.assertEqual(event2.originator_id, "1")
        self.assertEqual(event1.originator_version, 0)
        self.assertEqual(event2.originator_version, 1)

        # Check the events are not equal to each other, whilst being equal to
        # themselves.
        self.assertEqual(event1, event1)
        self.assertEqual(event2, event2)
        self.assertNotEqual(event1, event2)


class TestEventWithTimestampAndOriginatorID(TestCase):
    def test_one_subclass(self):
        # Check base class can be sub-classed.
        class Event(EventWithTimestamp, EventWithOriginatorID):
            pass

        # Check construction requires an ID.
        with self.assertRaises(TypeError):
            Event()

        # Get timestamp before events.
        time1 = decimaltimestamp()

        # Construct events.
        event1 = Event(originator_id="1")
        event2 = Event(originator_id="1")

        # Check the entity IDs.
        self.assertEqual(event1.originator_id, "1")
        self.assertEqual(event2.originator_id, "1")

        # Check the event timestamps.
        self.assertLessEqual(time1, event1.timestamp)
        self.assertLessEqual(event1.timestamp, event2.timestamp)
        self.assertLessEqual(event2.timestamp, decimaltimestamp())

        # Check the events are not equal to each other, whilst being equal to
        # themselves.
        self.assertEqual(event1, event1)
        self.assertEqual(event2, event2)
        self.assertNotEqual(event1, event2)

    def test_two_subclasses(self):
        # Check base class can be sub-classed.
        class Event(EventWithTimestamp, EventWithOriginatorID):
            pass

        class Event2(EventWithTimestamp, EventWithOriginatorID):
            pass

        # Check subclass can be instantiated with 'originator_id' parameter.
        ID1 = uuid4()
        ID2 = uuid4()
        VALUE1 = "a string"
        VALUE2 = "another string"
        event1 = Event(originator_id=ID1)
        self.assertEqual(event1.originator_id, ID1)

        # Check event has a domain event ID, and a timestamp.
        self.assertTrue(event1.timestamp)
        self.assertIsInstance(event1.timestamp, Decimal)

        # Check subclass can be instantiated with 'timestamp' parameter.
        DOMAIN_EVENT_ID1 = create_timesequenced_event_id()
        event2 = Event(originator_id=ID1, timestamp=DOMAIN_EVENT_ID1)
        self.assertEqual(event2.timestamp, DOMAIN_EVENT_ID1)

        # Check subclass can be instantiated with other parameters.
        event3 = Event(originator_id=ID1, an_attribute=VALUE1)
        self.assertEqual(event3.an_attribute, VALUE1)

        # Check the attribute value cannot be changed.
        with self.assertRaises(AttributeError):
            event3.an_attribute = VALUE2
        self.assertEqual(event3.an_attribute, VALUE1)

        # Check an event is equal to itself.
        self.assertTrue(event3 == event3)
        self.assertFalse(event3 != event3)
        event4 = Event(
            originator_id=ID1, timestamp=event3.timestamp, an_attribute=VALUE1
        )
        self.assertEqual(event3, event4)

        # Check domain events with same domain event ID have the same timestamp.
        event5 = Event(
            originator_id=event1.originator_id,
            timestamp=event1.timestamp,
            an_attribute=VALUE1,
        )
        self.assertEqual(event1.timestamp, event5.timestamp)

        # Check it's not equal to different type with same values.
        self.assertFalse(
            event3
            == Event2(
                originator_id=event3.originator_id, an_attribute=event3.an_attribute
            )
        )
        self.assertTrue(event3 != Event2(originator_id=ID1, an_attribute=VALUE1))

        # Check it's not equal to same type with different values.
        # - different originator_id
        self.assertNotEqual(event3.originator_id, ID2)
        self.assertEqual(event3.an_attribute, VALUE1)
        self.assertFalse(event2 == Event(originator_id=ID2, an_attribute=VALUE1))
        # - different attribute value
        self.assertTrue(event2 != Event(originator_id=ID1, an_attribute=VALUE2))

        # Check domain events with different domain event IDs have different timestamps.
        event4 = Event(originator_id=ID1, an_attribute=VALUE1)
        self.assertNotEqual(event2.timestamp, event4.timestamp)
        self.assertNotEqual(event2.timestamp, event4.timestamp)


# Todo: Review and reduce. This is the original test case, much but not all
#  of which is covered by the new tests above.
class TestEvents(TestCase):
    def tearDown(self):
        clear_event_handlers()

    def test_event_attributes(self):
        entity_id1 = uuid4()
        event = Example.Created(
            originator_id=entity_id1, originator_topic=get_topic(Example), a=1, b=2
        )

        # Check constructor keyword args lead to read-only attributes.
        self.assertEqual(1, event.a)
        self.assertEqual(2, event.b)
        self.assertRaises(AttributeError, getattr, event, "c")
        self.assertRaises(AttributeError, setattr, event, "c", 3)

        # Check domain event has auto-generated timestamp.
        self.assertIsInstance(event.timestamp, Decimal)

        # Check timestamp value can be given to domain events.
        event1 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1,
            b=2,
            timestamp=3,
        )
        self.assertEqual(3, event1.timestamp)

    def test_publish_subscribe_unsubscribe(self):
        # Check subscribing event handlers with predicates.
        # - when predicate is True, handler should be called
        event = mock.Mock()
        predicate = mock.Mock()
        handler = mock.Mock()

        # Check we can assert there are no event handlers subscribed.
        assert_event_handlers_empty()

        # When predicate is True, handler should be called ONCE.
        subscribe(handler=handler, predicate=predicate)

        # Check we can assert there are event handlers subscribed.
        self.assertRaises(EventHandlersNotEmptyError, assert_event_handlers_empty)

        # Check what happens when an event is published.
        publish([event])
        predicate.assert_called_once_with([event])
        handler.assert_called_once_with([event])

        # When predicate is True, after unsubscribing, handler should NOT be called
        # again.
        unsubscribe(handler=handler, predicate=predicate)
        publish([event])
        predicate.assert_called_once_with([event])
        handler.assert_called_once_with([event])

        # Check we can assert there are no event handlers subscribed.
        assert_event_handlers_empty()

        # When predicate is False, handler should NOT be called.
        predicate = lambda x: False
        handler = mock.Mock()
        subscribe(handler=handler, predicate=predicate)
        publish([event])
        self.assertEqual(0, handler.call_count)

        # Unsubscribe.
        unsubscribe(handler=handler, predicate=predicate)

        # Check we can assert there are no event handlers subscribed.
        assert_event_handlers_empty()

    def test_hash(self):
        entity_id1 = uuid4()
        event1 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1,
            b=2,
            timestamp=3,
        )
        event2 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1,
            b=2,
            timestamp=3,
        )
        self.assertEqual(hash(event1), hash(event2))

    def test_equality_comparison(self):
        entity_id1 = uuid4()
        event1 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1,
            b=2,
            timestamp=3,
        )
        event2 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1,
            b=2,
            timestamp=3,
        )
        event3 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=3,
            b=2,
            timestamp=3,
        )
        self.assertEqual(event1, event2)
        self.assertNotEqual(event1, event3)
        self.assertNotEqual(event2, event3)
        self.assertNotEqual(event2, None)

    def test_repr(self):
        entity_id1 = UUID("c7383095-85a3-4c49-ae32-9c9957ac8579")
        event1 = Example.Created(
            originator_id=entity_id1,
            originator_topic=get_topic(Example),
            a=1,
            b=2,
            timestamp=3,
        )
        self.maxDiff = None
        self.assertEqual(
            (
                "Example.Created("
                "__event_hash__='dbd0d45ed797de4f9a5d574ec4e3da3df74e67641bd7632dd05ffa28a5876391', "
                "__event_hash_method_name__='__hash_object_v2__', "
                "__event_topic__='eventsourcing.example.domainmodel#Example.Created', "
                "a=1, b=2, "
                "originator_id={}, "
                "originator_topic='eventsourcing.example.domainmodel#Example', "
                "originator_version=0, timestamp=3)"
            ).format(repr(entity_id1)),
            repr(event1),
        )

    def test_topic_resolution_error(self):
        # Check topic resolution error is raised, if the module path is
        # broken, and if the class name is broken.
        resolve_topic("eventsourcing.domain.model.events#DomainEvent")
        with self.assertRaises(TopicResolutionError):
            resolve_topic("eventsourcing.domain.model.broken#DomainEvent")
        with self.assertRaises(TopicResolutionError):
            resolve_topic("eventsourcing.domain.model.events#Broken")


class TestEventWithHash(TestCase):
    def test_event_hash_versioning(self):
        event = EventWithHash(a=1, b=1)
        self.assertTrue("__event_hash__", event.__dict__)
        self.assertIn("__event_hash_method_name__", event.__dict__)
        self.assertEqual(
            event.__dict__["__event_hash_method_name__"], "__hash_object_v2__"
        )
        event.__check_hash__()

        # Redo hashing, as a version 1, but with v2 __event_hash_method_name__.
        event.__dict__.pop("__event_hash__")
        event.__dict__["__event_hash__"] = event.__hash_object_v1__(event.__dict__)
        with self.assertRaises(EventHashError):
            event.__check_hash__()

        # Redo hashing, as a version 1, and without __event_hash_method_name__.
        event.__dict__.pop("__event_hash_method_name__")
        event.__dict__.pop("__event_hash__")
        event.__dict__["__event_hash__"] = event.__hash_object_v1__(event.__dict__)
        event.__check_hash__()

        # Redo hashing, as a version 2, but without __event_hash_method_name__.
        #
        # This should fail because it will fail to find an __event_hash_method_name__
        # and so it will assume the state wasn't recorded with that, and so use v1 by
        # default, which won't match with the v2 hash created below.
        event.__dict__.pop("__event_hash__")
        event.__dict__["__event_hash__"] = event.__hash_object_v2__(event.__dict__)
        with self.assertRaises(EventHashError):
            event.__check_hash__()



from uuid import uuid4

from axonclient.client import DEFAULT_LOCAL_AXONSERVER_URI, AxonClient, AxonEvent

from eventsourcing.infrastructure.axonserver.datastore import (
    AxonDatastore,
    AxonSettings,
)
from eventsourcing.infrastructure.axonserver.factory import AxonInfrastructureFactory
from eventsourcing.tests.datastore_tests import base


class AxonDatastoreTestCase(base.AbstractDatastoreTestCase):
    """
    Base class for test cases that use an Axon datastore.
    """

    infrastructure_factory_class = AxonInfrastructureFactory

    def construct_datastore(self):
        return AxonDatastore(settings=AxonSettings())

    def create_factory_kwargs(self):
        kwargs = super(AxonDatastoreTestCase, self).create_factory_kwargs()
        kwargs["axon_client"] = AxonClient(self.datastore.settings.uri)
        return kwargs


class TestAxonDatastore(AxonDatastoreTestCase, base.DatastoreTestCase):
    """
    Test case for Axon datastore.
    """

    def list_records(self):
        datastore = self.datastore
        assert isinstance(datastore, AxonDatastore)
        assert isinstance(datastore.axon_client, AxonClient)
        query = datastore.axon_client.list_events()
        return list(query)

    def create_record(self):
        datastore = self.datastore
        assert isinstance(datastore, AxonDatastore)
        assert isinstance(datastore.axon_client, AxonClient)
        datastore.axon_client.append_event(
            AxonEvent(
                message_identifier=str(uuid4()),
                aggregate_identifier=uuid4().hex,
                aggregate_sequence_number=0,
                aggregate_type="AggregateRoot",
                timestamp=0,
                payload_type="a",
                payload_revision="1",
                payload_data=b"",
                snapshot=False,
                meta_data={},
            )
        )



from datetime import datetime
from threading import Event, Lock

from eventsourcing.application.simple import Prompt


class RayDbJob(object):
    def __init__(self, method, args, kwargs):
        self.method = method
        self.args = args
        self.kwargs = kwargs
        self.is_done = Event()
        self.is_cancelled = Event()
        self.constructed = datetime.now()
        self.completed = None
        self.error = None
        self.time_lock = Lock()

    def __repr__(self):
        return "RayDbJob(method=%s)" % self.method

    def execute(self):
        self.started = datetime.now()
        try:
            self.result = self.method(*self.args, **self.kwargs)
        except Exception as e:
            self.error = e
            raise e
        finally:
            self.completed = datetime.now()
            self.is_done.set()

    def wait(self, timeout=None):
        return self.is_done.wait(timeout=timeout)

    @property
    def delay(self):
        if self.constructed and self.started:
            return self.started - self.constructed

    @property
    def duration(self):
        if self.completed and self.started:
            return self.completed - self.started


class RayPrompt(Prompt):
    def __init__(
        self,
        process_name: str,
        pipeline_id: int,
        head_notification_id=None,
        notification_ids=(),
        notifications=(),
    ):
        self.process_name: str = process_name
        self.pipeline_id: int = pipeline_id
        self.head_notification_id = head_notification_id
        self.notification_ids = notification_ids
        self.notifications = notifications

    def __repr__(self) -> str:
        return "{}({}={}, {}={}, {}={})".format(
            type(self).__name__,
            "process_name",
            self.process_name,
            "pipeline_id",
            self.pipeline_id,
            "head_notification_id",
            self.head_notification_id,
        )


class ProcessHasStopped(Exception):
    pass



from eventsourcing.tests.example_application_tests import base
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_record_manager import (
    SQLAlchemyRecordManagerTestCase,
)


class TestExampleApplicationWithSQLAlchemy(
    SQLAlchemyRecordManagerTestCase, base.ExampleApplicationTestCase
):
    pass



import datetime
from base64 import b64decode, b64encode
from collections import deque
from decimal import Decimal
from enum import Enum
from functools import singledispatch, wraps
from inspect import isfunction
from json import JSONDecoder, JSONEncoder
from modulefinder import Module
from types import FunctionType, MethodType
from typing import Optional
from uuid import UUID

import dateutil.parser

from eventsourcing.exceptions import EncoderTypeError
from eventsourcing.utils.topic import get_topic, resolve_topic

try:
    import orjson
except ImportError:
    orjson: Optional[Module] = None  # type: ignore


JSON_SEPARATORS = (",", ":")


def encoderpolicy(arg=None):
    """
    Decorator for encoder policy.

    Allows default behaviour to be built up from methods
    registered for different types of things, rather than
    chain of isinstance() calls in a long if-else block.
    """

    def _mutator(func):
        wrapped = singledispatch(func)

        @wraps(wrapped)
        def wrapper(*args, **kwargs):
            obj = kwargs.get("obj") or args[-1]
            return wrapped.dispatch(type(obj))(*args, **kwargs)

        wrapper.register = wrapped.register

        return wrapper

    assert isfunction(arg), arg
    return _mutator(arg)


def decoderpolicy(arg=None):
    """
    Decorator for decoder policy.

    Allows default behaviour to be built up from methods
    registered for different named keys, rather than
    chain of "in dict" queries in a long if-else block.
    """

    def _mutator(func):
        wrapped = func

        decoder_map = {}

        @wraps(wrapped)
        def wrapper(*args, **kwargs):

            d = kwargs.get("d") or args[-1]
            keys = list(d.keys())

            if len(keys) == 1:
                try:
                    decoder_func = decoder_map[keys[0]]
                except KeyError:
                    return d
                else:
                    return decoder_func(d)
            else:
                return d

        def register(key):
            def decorator(decoder_func):
                decoder_map[key] = decoder_func
                return decoder_func

            return decorator

        wrapper.register = register

        return wrapper

    assert isfunction(arg), arg
    return _mutator(arg)


class ObjectJSONEncoder(JSONEncoder):
    def __init__(self, sort_keys=False):
        super().__init__(sort_keys=sort_keys, separators=JSON_SEPARATORS)

    def encode(self, o):
        o = self.encode_object(o)
        if self.sort_keys is True or orjson is None:
            return super(ObjectJSONEncoder, self).encode(o).encode("utf8")
        else:
            return orjson.dumps(o)

    def encode_object(self, o):
        return self.encode_container(encoder(o))

    @encoderpolicy
    def encode_container(self, o):
        return o

    @encode_container.register(dict)
    def encode_dict(self, o):
        if type(o) == dict:
            return self.encode_dict_state(o)
        else:
            return {
                "__dict__": {
                    "topic": get_topic(o.__class__),
                    "state": self.encode_dict_state(o),
                }
            }

    def encode_dict_state(self, o):
        return {k: self.encode_object(v) for (k, v) in o.items()}

    @encode_container.register(tuple)
    def encode_tuple(self, o):
        if type(o) == tuple:
            return {"__tuple__": self.encode_object(list(o))}
        else:
            return {
                "__tuple__": {
                    "state": self.encode_object(list(o)),
                    "topic": get_topic(o.__class__),
                }
            }

    @encode_container.register(list)
    def encode_list(self, o):
        if type(o) == list:
            return [self.encode_object(i) for i in o]
        else:
            return {
                "__list__": {
                    "state": [self.encode_object(i) for i in o],
                    "topic": get_topic(o.__class__),
                }
            }

    @encode_container.register(set)
    def encode_set(self, o):
        if type(o) == set:
            return {"__set__": self.encode_iterable(o)}
        else:
            return {
                "__set__": {
                    "state": self.encode_iterable(o),
                    "topic": get_topic(o.__class__),
                }
            }

    def encode_iterable(self, o):
        return self.encode_object(self.sort_keys and sorted(o) or list(o))

    @encode_container.register(frozenset)
    def encode_frozenset(self, o):
        if type(o) == frozenset:
            return {"__frozenset__": self.encode_iterable(o)}
        else:
            return {
                "__frozenset__": {
                    "state": self.encode_iterable(o),
                    "topic": get_topic(o.__class__),
                }
            }

    @encode_container.register(deque)
    def encode_deque(self, o):
        if type(o) == deque:
            return {"__deque__": self.encode_object(list(o))}
        else:
            return {
                "__deque__": {
                    "state": self.encode_object(list(o)),
                    "topic": get_topic(o.__class__),
                }
            }

    @encode_container.register(object)
    def encode_instance(self, o):
        if hasattr(o, "__slots__") and o.__slots__ != ():
            topic = get_topic(o.__class__)
            state = {k: self.encode_object(getattr(o, k)) for k in o.__slots__}
            return {"__class__": {"state": state, "topic": topic}}
        elif hasattr(o, "__dict__"):
            topic = get_topic(o.__class__)
            state = {k: self.encode_object(v) for k, v in o.__dict__.items()}
            return {"__class__": {"state": state, "topic": topic}}
        else:
            return o


@encoderpolicy
def encoder(o):
    return o


class ObjectJSONDecoder(JSONDecoder):
    def __init__(self, object_hook=None, **kwargs):
        super(ObjectJSONDecoder, self).__init__(
            object_hook=object_hook or decoder, **kwargs
        )


@decoderpolicy
def decoder(d):
    return d


@encoder.register(type)
def encode_type(o):
    return {"__type__": get_topic(o)}


@encoder.register(MethodType)
def encode_method(o):
    raise EncoderTypeError(o)


@encoder.register(FunctionType)
def encode_function(o):
    raise EncoderTypeError(o)


@decoder.register("__type__")
def decode_type(d):
    return resolve_topic(d["__type__"])


@decoder.register("__class__")
def decode_object(d):
    topic = d["__class__"]["topic"]
    state = d["__class__"]["state"]
    obj_class = resolve_topic(topic)
    obj = object.__new__(obj_class)
    if hasattr(obj, "__dict__"):
        obj.__dict__.update(state)
    else:
        for k, v in state.items():
            object.__setattr__(obj, k, v)
    return obj


@encoder.register(UUID)
def encode_uuid(obj):
    return {"UUID": obj.hex}


@decoder.register("UUID")
def decode_uuid(d):
    return UUID(d["UUID"])


@encoder.register(datetime.datetime)
def encode_datetime(obj):
    return {"ISO8601_datetime": obj.strftime("%Y-%m-%dT%H:%M:%S.%f%z")}


@decoder.register("ISO8601_datetime")
def decode_datetime(d):
    return dateutil.parser.parse(d["ISO8601_datetime"])


@encoder.register(datetime.date)
def encode_date(obj):
    return {"ISO8601_date": obj.isoformat()}


@decoder.register("ISO8601_date")
def decode_date(d):
    return datetime.datetime.strptime(d["ISO8601_date"], "%Y-%m-%d").date()


@encoder.register(datetime.time)
def encode_time(obj):
    return {"ISO8601_time": obj.strftime("%H:%M:%S.%f")}


@decoder.register("ISO8601_time")
def decode_time(d):
    hour, minute, seconds = d["ISO8601_time"].split(":")
    second, microsecond = seconds.split(".")
    return datetime.time(int(hour), int(minute), int(second), int(microsecond))


@encoder.register(Decimal)
def encode_decimal(obj):
    return {"__decimal__": str(obj)}


@decoder.register("__decimal__")
def decode_decimal(d):
    return Decimal(d["__decimal__"])


@encoder.register(Enum)
def encode_enum(obj):
    return {"__enum__": {"topic": get_topic(type(obj)), "name": obj.name}}


@decoder.register("__enum__")
def decode_enum(d):
    topic = d["__enum__"]["topic"]
    name = d["__enum__"]["name"]
    enum = resolve_topic(topic)
    return getattr(enum, name)


@decoder.register("__deque__")
def decode_deque(d):
    deque_data = d["__deque__"]
    if type(deque_data) == dict:
        topic = deque_data["topic"]
        try:
            state = deque_data["state"]
        except KeyError:
            state = deque_data["values"]

        deque_type = resolve_topic(topic)
        return deque_type(state)
    else:
        return deque(deque_data)


@decoder.register("__tuple__")
def decode_tuple(d):
    tuple_data = d["__tuple__"]
    if type(tuple_data) == dict:
        # For NamedTuple objects.
        topic = tuple_data["topic"]
        state = tuple_data["state"]
        tuple_type = resolve_topic(topic)
        obj = tuple_type(*state)
    else:
        # For standard tuple objects.
        obj = tuple(tuple_data)
    return obj


@decoder.register("__dict__")
def decode_dict(d):
    topic = d["__dict__"]["topic"]
    state = d["__dict__"]["state"]
    dict_type = resolve_topic(topic)
    return dict_type(state)


@decoder.register("__set__")
def decode_set(d):
    set_data = d["__set__"]
    if isinstance(set_data, dict):
        topic = set_data["topic"]
        state = set_data["state"]
        set_type = resolve_topic(topic)
        return set_type(state)
    else:
        return set(set_data)


@decoder.register("__frozenset__")
def decode_frozenset(d):
    set_data = d["__frozenset__"]
    if isinstance(set_data, dict):
        topic = set_data["topic"]
        state = set_data["state"]
        set_type = resolve_topic(topic)
        return set_type(state)
    else:
        return frozenset(set_data)


@encoder.register(bytes)
def encode_bytes(o):
    return {"__bytes__": b64str_from_bytes(o)}


@decoder.register("__bytes__")
def decode_bytes(d):
    return bytes_from_b64str(d["__bytes__"])


def b64str_from_bytes(value: bytes) -> str:
    return b64encode(value).decode("utf8")


def bytes_from_b64str(value):
    return b64decode(value.encode("utf8"))



from functools import reduce
from threading import Lock
from typing import Any, Callable, Dict, Iterable, Optional, Type
from uuid import UUID

from eventsourcing.domain.model.entity import TVersionedEntity, TVersionedEvent
from eventsourcing.domain.model.events import AbstractSnapshot
from eventsourcing.domain.model.repository import AbstractEntityRepository
from eventsourcing.exceptions import RepositoryKeyError
from eventsourcing.infrastructure.base import AbstractEventStore, AbstractRecordManager
from eventsourcing.infrastructure.snapshotting import AbstractSnapshotStrategy
from eventsourcing.whitehead import SEntity


class EventSourcedRepository(
    AbstractEntityRepository[TVersionedEntity, TVersionedEvent]
):
    # The page size by which events are retrieved. If this
    # value is set to a positive integer, the events of
    # the entity will be retrieved in pages, using a series
    # of queries, rather than with one potentially large query.
    __page_size__: Optional[int] = None

    def __init__(
        self,
        event_store: AbstractEventStore,
        use_cache: bool = False,
        snapshot_strategy: Optional[AbstractSnapshotStrategy] = None,
        mutator_func: Optional[
            Callable[
                [Optional[TVersionedEntity], TVersionedEvent],
                Optional[TVersionedEntity],
            ]
        ] = None,
        **kwargs: Any
    ):
        self._event_store: AbstractEventStore = event_store
        self._snapshot_strategy = snapshot_strategy
        self._mutator_func = mutator_func or self.mutate
        super(EventSourcedRepository, self).__init__()

        # NB If you use the cache, make sure to del entities
        # when records fail to write otherwise the cache will
        # give an entity that is ahead of the event records,
        # and writing more records will give a broken sequence.
        self._cache: Dict[UUID, Optional[TVersionedEntity]] = {}
        self._cache_lock = Lock()
        self._use_cache = use_cache

    @property
    def event_store(self) -> AbstractEventStore[TVersionedEvent, AbstractRecordManager]:
        """
        Returns event store object used by this repository.
        """
        return self._event_store

    @property
    def use_cache(self) -> bool:
        return self._use_cache

    @property
    def cache_lock(self) -> Lock:
        return self._cache_lock

    @property
    def cache(self) -> Dict:
        return self._cache

    @use_cache.setter
    def use_cache(self, value: bool) -> None:
        self._use_cache = value
        if not self._use_cache:
            self._cache.clear()

    def __contains__(self, entity_id: UUID) -> bool:
        """
        Returns a boolean value according to whether entity with given ID exists.
        """
        return self.get_entity(entity_id) is not None

    def __getitem__(self, entity_id: UUID) -> TVersionedEntity:
        """
        Returns entity with given ID.

        :param entity_id: ID of entity in the repository.
        :raises RepositoryKeyError: If the entity is not found.
        """
        if self._use_cache:
            try:
                # Get entity from the cache.
                with self._cache_lock:
                    entity: Optional[TVersionedEntity] = self._cache[entity_id]
            except KeyError:
                # Reconstitute the entity.
                entity = self.get_entity(entity_id)
                # Put entity in the cache.
                self.put_entity_in_cache(entity_id, entity)
        else:
            entity = self.get_entity(entity_id)

        # Never created or already discarded?
        if entity is None:
            raise RepositoryKeyError(entity_id)

        # Return entity.
        assert entity is not None
        return entity

    def put_entity_in_cache(self, entity_id: UUID, entity: TVersionedEntity):
        if entity is None or entity_id in self._cache:
            return
        with self._cache_lock:
            self._cache[entity_id] = entity

    def get_entity(
        self, entity_id: UUID, at: Optional[int] = None
    ) -> Optional[TVersionedEntity]:
        """
        Returns entity with given ID, optionally at a version.

        Returns None if entity not found.
        """

        # Get a snapshot (None if none exist).
        if (
            self._snapshot_strategy
            and not self.event_store.record_manager.has_integrated_snapshots
        ):
            snapshot = self._snapshot_strategy.get_snapshot(entity_id, lte=at)
        else:
            snapshot = None

        # Decide the initial state of the entity, and the
        # version of the last item applied to the entity.
        if snapshot is None:
            initial_state = None
            gt = None
        else:
            assert isinstance(snapshot, AbstractSnapshot), snapshot
            initial_state = snapshot.__mutate__(None)
            gt = snapshot.originator_version

        # Obtain and return current state.
        return self.get_and_project_events(
            entity_id, gt=gt, lte=at, initial_state=initial_state
        )

    def get_and_project_events(
        self,
        entity_id: UUID,
        gt: Optional[int] = None,
        gte: Optional[int] = None,
        lt: Optional[int] = None,
        lte: Optional[int] = None,
        limit: Optional[int] = None,
        initial_state: Optional[TVersionedEntity] = None,
        query_descending: bool = False,
    ) -> Optional[TVersionedEntity]:
        """
        Reconstitutes requested domain entity from domain events found in event store.
        """
        # Decide if query is in ascending order.
        #  - A "speed up" for when events are stored in descending order (e.g.
        #  in Cassandra) and it is faster to get them in that order.
        #  - This isn't useful when 'until' or 'after' or 'limit' are set,
        #    because the inclusiveness or exclusiveness of until and after
        #    and the end of the stream that is truncated by limit both depend on
        #    the direction of the query. Also paging backwards isn't useful, because
        #    all the events are needed eventually, so it would probably slow things
        #    down. Paging is intended to support replaying longer event streams, and
        #    only makes sense to work in ascending order.
        if (
            gt is None
            and gte is None
            and lt is None
            and lte is None
            and self.__page_size__ is None
        ):
            is_ascending = True
        else:
            is_ascending = not query_descending

        # Get entity's domain events from the event store.
        domain_events = self.event_store.iter_events(
            originator_id=entity_id,
            gt=gt,
            gte=gte,
            lt=lt,
            lte=lte,
            limit=limit,
            is_ascending=is_ascending,
            page_size=self.__page_size__,
        )

        # The events must be replayed in ascending order.
        if not is_ascending:
            domain_events = reversed(list(domain_events))

        # Project the domain events onto the initial state.
        return self.project_events(initial_state, domain_events)

    def project_events(
        self,
        initial_state: Optional[TVersionedEntity],
        domain_events: Iterable[TVersionedEvent],
    ) -> Optional[TVersionedEntity]:
        """
        Evolves initial_state using the domain_events and a mutator function.

        Applies a mutator function cumulatively to a sequence of domain
        events, so as to mutate the initial value to a mutated value.

        This class's mutate() method is used as the default mutator function, but
        custom behaviour can be introduced by passing in a 'mutator_func' argument
        when constructing this class, or by overridding the mutate() method.
        """
        return reduce(self._mutator_func, domain_events, initial_state)

    @staticmethod
    def mutate(
        initial: Optional[TVersionedEntity], event: TVersionedEvent
    ) -> Optional[TVersionedEntity]:
        """
        Default mutator function, which uses __mutate__()
        method on event object to mutate initial state.

        :param initial: Initial state to be mutated by this function.
        :param event: Event that causes the initial state to be mutated.
        :return: Returns the mutated state.
        """
        # Check obj is not None.
        if initial is not None:
            event.__check_obj__(initial)
        return event.__mutate__(initial)

    # Todo: Does this method belong on this class?
    def take_snapshot(
        self, entity_id: UUID, lt: Optional[int] = None, lte: Optional[int] = None
    ) -> Optional[AbstractSnapshot]:
        """
        Takes a snapshot of the entity as it existed after the most recent
        event, optionally less than, or less than or equal to, a particular position.
        """
        snapshot = None
        if self._snapshot_strategy:
            if self.event_store.record_manager.has_integrated_snapshots:
                # Get entity.
                entity = self.get_and_project_events(entity_id=entity_id)

                # Take snapshot from entity.
                if entity is not None:
                    snapshot = self._snapshot_strategy.take_snapshot(
                        entity_id, entity, entity.__version__
                    )

            else:
                # Get the latest event (optionally until a particular position).
                latest_event = self.event_store.get_most_recent_event(
                    entity_id, lt=lt, lte=lte
                )

                # If there is something to snapshot, then look for a snapshot
                # taken before or at the entity version of the latest event. Please
                # note, the snapshot might have a smaller version number than
                # the latest event if events occurred since the latest snapshot was taken.
                if latest_event is not None:
                    latest_snapshot = self._snapshot_strategy.get_snapshot(
                        entity_id, lt=lt, lte=lte
                    )
                    latest_version = latest_event.originator_version

                    if (
                        latest_snapshot
                        and latest_snapshot.originator_version == latest_version
                    ):
                        # If up-to-date snapshot exists, there's nothing to do.
                        snapshot = latest_snapshot
                    else:
                        # Otherwise recover entity state from latest snapshot.
                        if latest_snapshot:
                            initial_state = latest_snapshot.__mutate__(None)
                            gt: Optional[int] = latest_snapshot.originator_version
                        else:
                            initial_state = None
                            gt = None

                        # Fast-forward entity state to latest version.
                        entity = self.get_and_project_events(
                            entity_id=entity_id,
                            gt=gt,
                            lte=latest_version,
                            initial_state=initial_state,
                        )

                        # Take snapshot from entity.
                        snapshot = self._snapshot_strategy.take_snapshot(
                            entity_id, entity, latest_version
                        )

        return snapshot

    def get_instance_of(
        self, instance_class: Type[SEntity], entity_id: UUID, at: Optional[int] = None
    ) -> Optional[SEntity]:
        entity = self.get_entity(entity_id, at=at)
        if isinstance(entity, instance_class):
            return entity
        else:
            return None



from typing import Any

from eventsourcing.application.simple import ApplicationWithConcreteInfrastructure
from eventsourcing.infrastructure.django.factory import DjangoInfrastructureFactory
from eventsourcing.infrastructure.django.utils import (
    close_django_connection,
    setup_django,
)


class DjangoApplication(ApplicationWithConcreteInfrastructure):
    infrastructure_factory_class = DjangoInfrastructureFactory

    def __init__(self, tracking_record_class: Any = None, *args: Any, **kwargs: Any):
        self._tracking_record_class = tracking_record_class
        super(DjangoApplication, self).__init__(*args, **kwargs)

    @property
    def stored_event_record_class(self) -> type:  # type: ignore
        # This is awkward, but need to avoid importing library Django models.
        from eventsourcing.infrastructure.django.models import StoredEventRecord

        return StoredEventRecord

    @property
    def snapshot_record_class(cls) -> type:  # type: ignore
        # This is awkward, but need to avoid importing library Django models.
        from eventsourcing.infrastructure.django.models import EntitySnapshotRecord

        return EntitySnapshotRecord

    @property
    def tracking_record_class(cls) -> Any:
        from eventsourcing.infrastructure.django.models import (
            NotificationTrackingRecord,
        )

        return NotificationTrackingRecord

    def construct_infrastructure(self, *args: Any, **kwargs: Any) -> None:
        tracking_record_class = (
            self._tracking_record_class or self.tracking_record_class
        )
        super(DjangoApplication, self).construct_infrastructure(
            tracking_record_class=tracking_record_class, *args, **kwargs
        )

    @classmethod
    def reset_connection_after_forking(cls) -> None:
        """
        Resets database connection after forking.
        """
        close_django_connection()
        setup_django()



# from eventsourcing.exceptions import RepositoryKeyError
#
#
# def get_string_ids(node_id, node_repo, node_child_collection_repo, stringid_collection_repo, length_until_end=0,
#                    edge_length=0,
#                    limit=None, hop_count=0, hop_max=None):
#     """Generator that yields unique string IDs from leaf nodes as they are discovered,
#     by performing a depth first search on the suffix tree from the given node.
#     """
#     stack = list()
#
#     # _print("Searching for string IDs from node: {}".format(node_id))
#
#     stack.append((node_id, edge_length, None))
#     unique_node_ids = set(node_id)
#     unique_string_ids = set()
#     cum_lengths_until_end = {None: length_until_end}
#     while stack:
#
#         (node_id, edge_length, parent_node_id) = stack.pop()
#
#         length_until_end = cum_lengths_until_end[parent_node_id] + edge_length
#         cum_lengths_until_end[node_id] = length_until_end
#
#         node = node_repo[node_id]
#         assert isinstance(node, SuffixTreeNode)
#
#         try:
#             stringid_collection = stringid_collection_repo[node_id]
#         except KeyError:
#             stringid_collection = None
#         else:
#             assert isinstance(stringid_collection, StringidCollection)
#
#         if stringid_collection and stringid_collection.items:
#
#             # Deduplicate string IDs (the substring might match more than one suffix in any given string).
#             for string_id in stringid_collection.items:
#                 if string_id in unique_string_ids:
#                     continue
#
#                 # Check the match doesn't encroach upon the string's extension.
#                 extension_length = len(STRING_ID_END)
#                 if length_until_end < extension_length:
#                     continue
#
#                 # print("Found string ID '{}' at node: {}".format(string_id, node_id))
#
#                 # Remember the string ID, we only want one node per string ID.
#                 unique_string_ids.add(string_id)
#
#                 # Yield the string ID.
#                 yield string_id
#
#                 # Check the limit, stop if we've had enough string IDs.
#                 if limit is not None and len(unique_string_ids) >= limit:
#                     return
#
#         # Check the hop count before getting the child collection.
#         if hop_max is None or hop_count < hop_max:
#             try:
#                 node_child_collection = node_child_collection_repo[node_id]
#             except RepositoryKeyError:
#                 # It doesn't matter if there isn't a child collection.
#                 pass
#             else:
#                 # Since there's a collection, put the child nodes on the stack.
#                 child_node_ids = node_child_collection._child_node_ids.copy()
#                 for (child_node_id, edge_length) in child_node_ids.items():
#                     # Check the hop count, stop if we've seen enough tree nodes.
#                     #  - this works by choking off the stack, but we continue
#                     #    to process everything on the stack after this condition
#                     #    which saves putting loads of things on the stack that
#                     #    won't be used, and walking more tree than is necessary
#                     if hop_max is not None and hop_count >= hop_max:
#                         break
#
#                     # Put the child node on the stack.
#                     # if child_node_id not in unique_node_ids:
#                     if child_node_id != node_id:
#                         unique_node_ids.add(child_node_id)
#                         stack.append((child_node_id, edge_length, node.id))
#
#                     # Increase the hop count.
#                     hop_count += 1
#
#
# def find_substring_edge(substring, suffix_tree, edge_repo):
#     """Returns the last edge, if substring in tree, otherwise None.
#     """
#     assert isinstance(substring, str)
#     assert isinstance(suffix_tree, GeneralizedSuffixTree)
#     assert isinstance(edge_repo, EdgeRepository)
#     if not substring:
#         return None, None
#     if suffix_tree.case_insensitive:
#         substring = substring.lower()
#     curr_node_id = suffix_tree.root_node_id
#     i = 0
#     while i < len(substring):
#         edge_id = make_edge_id(curr_node_id, substring[i])
#         try:
#             edge = edge_repo[edge_id]
#         except RepositoryKeyError:
#             return None, None
#         ln = min(edge.length + 1, len(substring) - i)
#         if substring[i:i + ln] != edge.label[:ln]:
#             return None, None
#         i += edge.length + 1
#         curr_node_id = edge.dest_node_id
#     return edge, ln
#
#
# def has_substring(substring, suffix_tree, edge_repo):
#     edge, ln = find_substring_edge(substring, suffix_tree, edge_repo)
#     return edge is not None



from Crypto.Cipher import AES

from eventsourcing.exceptions import DataIntegrityError
from eventsourcing.utils.random import random_bytes


class AESCipher(object):
    """
    Cipher strategy that uses Crypto library AES cipher in GCM mode.
    """

    def __init__(self, cipher_key: bytes):
        """
        Initialises AES cipher strategy with ``cipher_key``.

        :param cipher_key: 16, 24, or 32 random bytes
        """
        assert len(cipher_key) in [16, 24, 32]
        self.cipher_key = cipher_key

    def encrypt(self, plaintext: bytes) -> bytes:
        """Return ciphertext for given plaintext."""

        # Construct AES-GCM cipher, with 96-bit nonce.
        cipher = AES.new(self.cipher_key, AES.MODE_GCM, nonce=random_bytes(12))

        # Encrypt and digest.
        encrypted, tag = cipher.encrypt_and_digest(plaintext)  # type: ignore

        # Combine with nonce.
        ciphertext = cipher.nonce + tag + encrypted  # type: ignore

        # Return ciphertext.
        return ciphertext

    def decrypt(self, ciphertext: bytes) -> bytes:
        """Return plaintext for given ciphertext."""

        # Split out the nonce, tag, and encrypted data.
        nonce = ciphertext[:12]
        if len(nonce) != 12:
            raise DataIntegrityError("Cipher text is damaged: invalid nonce length")

        tag = ciphertext[12:28]
        if len(tag) != 16:
            raise DataIntegrityError("Cipher text is damaged: invalid tag length")

        encrypted = ciphertext[28:]

        # Construct AES cipher, with old nonce.
        cipher = AES.new(self.cipher_key, AES.MODE_GCM, nonce)

        # Decrypt and verify.
        try:
            plaintext = cipher.decrypt_and_verify(encrypted, tag)  # type: ignore
        except ValueError as e:
            raise DataIntegrityError("Cipher text is damaged: {}".format(e))
        return plaintext



from eventsourcing.infrastructure.cassandra.manager import CassandraRecordManager
from eventsourcing.infrastructure.cassandra.records import (
    IntegerSequencedRecord,
    SnapshotRecord,
    TimestampSequencedRecord,
)
from eventsourcing.infrastructure.factory import InfrastructureFactory


class CassandraInfrastructureFactory(InfrastructureFactory):
    """
    Infrastructure factory for Cassandra.
    """

    record_manager_class = CassandraRecordManager
    integer_sequenced_record_class = IntegerSequencedRecord
    timestamp_sequenced_record_class = TimestampSequencedRecord
    snapshot_record_class = SnapshotRecord



import hashlib
import os

from eventsourcing.utils.transcoding import ObjectJSONEncoder

SALT_FOR_DATA_INTEGRITY = os.getenv("SALT_FOR_DATA_INTEGRITY", "")


def hash_object(json_encoder: ObjectJSONEncoder, obj: dict) -> str:
    """
    Calculates SHA-256 hash of JSON encoded 'obj'.

    :param json_encoder: JSON encoder object.
    :param obj: Object to be hashed.
    :return: SHA-256 as hexadecimal string.
    :rtype str
    """
    s = json_encoder.encode((obj, SALT_FOR_DATA_INTEGRITY))
    return hashlib.sha256(s).hexdigest()



import os
from typing import Any, Dict, Optional, Sequence, Union

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.exc import InternalError
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm import Session, scoped_session, sessionmaker
from sqlalchemy.pool import StaticPool

from eventsourcing.infrastructure.datastore import AbstractDatastore, DatastoreSettings
from eventsourcing.infrastructure.sqlalchemy.records import Base

SQLITE_IN_MEMORY = "sqlite:///:memory:"
DEFAULT_SQLALCHEMY_DB_URI = SQLITE_IN_MEMORY
# DEFAULT_SQLALCHEMY_DB_URI = 'sqlite:///FILE_SYSTEM_PATH'
# DEFAULT_SQLALCHEMY_DB_URI = 'mysql://username:password@localhost/eventsourcing'
# DEFAULT_SQLALCHEMY_DB_URI = 'postgresql://username:password@localhost:5432/eventsourcing'
DEFAULT_SQLALCHEMY_DB_POOL_SIZE = 5


class SQLAlchemySettings(DatastoreSettings):
    def __init__(self, uri: Optional[str] = None, pool_size: Optional[int] = None):
        if uri is not None:
            self.uri = uri
        else:
            self.uri = os.getenv("DB_URI", DEFAULT_SQLALCHEMY_DB_URI)

        if pool_size is not None:
            self.pool_size = pool_size
        else:
            self.pool_size = int(
                os.getenv("DB_POOL_SIZE", DEFAULT_SQLALCHEMY_DB_POOL_SIZE)
            )


class SQLAlchemyDatastore(AbstractDatastore[SQLAlchemySettings]):
    def __init__(
        self,
        settings: SQLAlchemySettings,
        base: DeclarativeMeta = Base,
        tables: Optional[Sequence] = None,
        connection_strategy: str = "plain",
        session: Optional[Union[Session, scoped_session]] = None,
    ):
        super(SQLAlchemyDatastore, self).__init__(settings=settings)
        self._was_session_created_here = False
        self._session = session
        if session:
            self._engine: Optional[Engine] = session.get_bind()
        else:
            self._engine = None
        self._base = base
        self._tables = tables
        self._connection_strategy = connection_strategy

    @property
    def session(self) -> Union[Session, scoped_session]:
        if self._session is None:
            if self._engine is None:
                self.setup_connection()
            session_factory = sessionmaker(bind=self._engine)
            self._session = scoped_session(session_factory)
            self._was_session_created_here = True
        return self._session

    def setup_connection(self) -> None:
        assert isinstance(self.settings, SQLAlchemySettings), self.settings
        if self._engine is None:

            # Create SQLAlchemy engine.
            if self.is_sqlite():
                kwargs: Dict[str, Any] = {"connect_args": {"check_same_thread": False}}
            elif self.settings.pool_size == 1:
                kwargs = {"poolclass": StaticPool}
            else:
                kwargs = {"pool_size": self.settings.pool_size}

            self._engine = create_engine(
                self.settings.uri,
                strategy=self._connection_strategy,
                # echo=True,
                **kwargs
            )
            assert self._engine

    def is_sqlite(self) -> bool:
        return self.settings.uri is not None and self.settings.uri.startswith("sqlite")

    def setup_tables(self) -> None:
        if self._tables is not None:
            for table in self._tables:
                self.setup_table(table)

    def setup_table(self, table: Any) -> None:
        if self._engine is None:
            raise Exception("Engine not set when required: {}".format(self))
        try:
            table.__table__.create(self._engine, checkfirst=True)
        except InternalError as e:
            if "Table '{}' already exists".format(table.__tablename__) in str(e):
                # This is a race condition from checkfirst=True. Can happen
                # if two threads call this method at the same time.
                pass
            else:
                raise

    def drop_tables(self) -> None:
        if self._tables is not None:
            for table in self._tables:
                self.drop_table(table)

    def drop_table(self, table: Any) -> None:
        table.__table__.drop(self._engine, checkfirst=True)

    def truncate_tables(self) -> None:
        self.drop_tables()

    def close_connection(self) -> None:
        if self._was_session_created_here:
            if self._session:
                self._session.close()
                self._session = None
            if self._engine:
                # Call dispose(), unless sqlite (to avoid error 'stored_events'
                # table does not exist in projections.rst doc).
                if not self.is_sqlite():
                    self._engine.dispose()
                self._engine = None



from typing import Any, Optional

from eventsourcing.application.policies import SnapshottingPolicy
from eventsourcing.application.simple import SimpleApplication
from eventsourcing.domain.model.entity import TVersionedEntity, TVersionedEvent
from eventsourcing.domain.model.events import AbstractSnapshot
from eventsourcing.infrastructure.base import AbstractEventStore, AbstractRecordManager
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.snapshotting import EventSourcedSnapshotStrategy


class SnapshottingApplication(SimpleApplication[TVersionedEntity, TVersionedEvent]):
    snapshot_period = 0

    def __init__(self, snapshot_period: int = 0, **kwargs: Any):
        self.snapshot_period = snapshot_period or self.snapshot_period
        self.snapshot_store: Optional[
            AbstractEventStore[AbstractSnapshot, AbstractRecordManager]
        ] = None
        self.snapshot_strategy: Optional[EventSourcedSnapshotStrategy] = None
        self.snapshotting_policy: Optional[SnapshottingPolicy] = None
        super(SnapshottingApplication, self).__init__(**kwargs)

    def construct_event_store(self) -> None:
        super(SnapshottingApplication, self).construct_event_store()
        # Setup event store for snapshots.
        assert self.infrastructure_factory
        record_manager = self.infrastructure_factory.construct_snapshot_record_manager()
        assert self.sequenced_item_mapper_class is not None
        assert self.sequenced_item_class is not None

        self.snapshot_store = EventStore(
            record_manager=record_manager, event_mapper=self.event_store.event_mapper
        )

    def construct_repository(self, **kwargs: Any) -> None:
        # Setup repository with a snapshot strategy.
        assert self.snapshot_store
        self.snapshot_strategy = EventSourcedSnapshotStrategy(
            snapshot_store=self.snapshot_store
        )
        super(SnapshottingApplication, self).construct_repository(
            snapshot_strategy=self.snapshot_strategy, **kwargs
        )

    def construct_persistence_policy(self) -> None:
        super(SnapshottingApplication, self).construct_persistence_policy()
        assert self.snapshot_store
        self.snapshotting_policy = SnapshottingPolicy(
            repository=self.repository,
            snapshot_store=self.snapshot_store,
            persist_event_type=self.persist_event_type,
            period=self.snapshot_period,
        )

    def setup_table(self) -> None:
        super(SnapshottingApplication, self).setup_table()
        if self._datastore is not None:
            assert self.snapshot_store
            self._datastore.setup_table(self.snapshot_store.record_manager.record_class)

    def drop_table(self) -> None:
        super(SnapshottingApplication, self).drop_table()
        if self._datastore is not None:
            assert self.snapshot_store
            self._datastore.drop_table(self.snapshot_store.record_manager.record_class)

    def close(self) -> None:
        super(SnapshottingApplication, self).close()
        if self.snapshotting_policy is not None:
            self.snapshotting_policy.close()
            self.snapshotting_policy = None



from typing import Generic, Optional, Tuple, Union

from eventsourcing.domain.model.entity import VersionedEntity
from eventsourcing.domain.model.events import (
    AbstractSnapshot,
    EventWithOriginatorVersion,
    subscribe,
    unsubscribe,
)
from eventsourcing.domain.model.repository import AbstractEntityRepository
from eventsourcing.infrastructure.base import AbstractEventStore, AbstractRecordManager
from eventsourcing.whitehead import IterableOfEvents, TEvent


class PersistencePolicy(object):
    """
    Stores events of given type to given event store, whenever they are published.
    """

    def __init__(
        self,
        event_store: AbstractEventStore,
        persist_event_type: Optional[Union[type, Tuple]] = None,
    ):
        self.event_store = event_store
        self.persist_event_type = persist_event_type
        subscribe(self.store_events, self.is_event)

    def close(self) -> None:
        unsubscribe(self.store_events, self.is_event)

    def is_event(self, events: IterableOfEvents) -> bool:
        if self.persist_event_type is None:
            return False
        elif type(events) not in [list, tuple]:
            return False
        else:
            return all(isinstance(e, self.persist_event_type) for e in events)

    def store_events(self, events: IterableOfEvents) -> None:
        self.event_store.store_events(events)


# Todo: Separate PeriodicSnapshottingPolicy from base class? Make usage more
#  configurable.
class SnapshottingPolicy(Generic[TEvent]):
    def __init__(
        self,
        repository: AbstractEntityRepository,
        snapshot_store: AbstractEventStore[AbstractSnapshot, AbstractRecordManager],
        persist_event_type: Optional[Union[type, Tuple]] = (
            EventWithOriginatorVersion,
        ),
        period: int = 0,
    ):
        self.repository = repository
        self.snapshot_store = snapshot_store
        self.period = period
        self.persist_event_type = persist_event_type
        subscribe(predicate=self.condition, handler=self.take_snapshot)

    def close(self) -> None:
        unsubscribe(predicate=self.condition, handler=self.take_snapshot)

    def condition(self, event: IterableOfEvents) -> bool:
        # Periodically by default.
        if self.persist_event_type and isinstance(self.period, int) and self.period > 0:
            if isinstance(event, (list, tuple)):
                for e in event:
                    if self.condition(e):
                        return True
            else:
                if isinstance(event, self.persist_event_type):
                    if isinstance(event, VersionedEntity.Event):
                        return (event.originator_version + 1) % self.period == 0
        return False

    def take_snapshot(self, events: IterableOfEvents) -> None:
        event = list(events)[-1]  # snapshot at the last version
        assert isinstance(event, VersionedEntity.Event), type(event)
        self.repository.take_snapshot(event.originator_id, lte=event.originator_version)



import itertools
from copy import deepcopy
from typing import Any, Dict, Optional, Sequence, Set, Union
from uuid import UUID

from eventsourcing.application.process import ProcessApplication, WrappedRepository
from eventsourcing.contrib.paxos.composable import (
    PaxosInstance,
    PaxosMessage,
    ProposalStatus,
    Resolution,
)
from eventsourcing.domain.model.aggregate import BaseAggregateRoot
from eventsourcing.domain.model.decorators import retry
from eventsourcing.exceptions import (
    OperationalError,
    RecordConflictError,
    RepositoryKeyError,
)
from eventsourcing.system.definition import System


class PaxosAggregate(BaseAggregateRoot):
    """
    Event-sourced Paxos participant.
    """

    paxos_variables = [
        "proposed_value",
        "proposal_id",
        "promised_id",
        "accepted_id",
        "accepted_value",
        "highest_proposal_id",
        "promises_received",
        "nacks_received",
        "highest_accepted_id",
        "leader",
        "proposals",
        "acceptors",
        "final_value",
    ]
    is_verbose = False

    def __init__(self, quorum_size: int, network_uid: str, **kwargs: Any):
        assert isinstance(quorum_size, int)
        self.quorum_size = quorum_size
        self.network_uid = network_uid
        self.promises_received: Set[str] = set()
        self.nacks_received: Set[str] = set()
        self.leader = False
        self.proposals: Dict[str, ProposalStatus] = {}
        self.acceptors: Dict[str, str] = {}
        self.final_value = None
        super(PaxosAggregate, self).__init__(**kwargs)

    @property
    def paxos_instance(self) -> PaxosInstance:
        """
        Returns instance of PaxosInstance (protocol implementation).
        """
        # Construct instance with the constant attributes.
        instance = PaxosInstance(self.network_uid, self.quorum_size)

        # Set the variable attributes from the aggregate.
        for name in self.paxos_variables:
            value = getattr(self, name, None)
            if value is not None:
                if isinstance(value, (set, list, dict, tuple)):
                    value = deepcopy(value)
                setattr(instance, name, value)

        # Return the instance.
        return instance

    class Event(BaseAggregateRoot.Event["PaxosAggregate"]):
        """
        Base event class for PaxosAggregate.
        """

    class Started(Event, BaseAggregateRoot.Created["PaxosAggregate"]):
        """
        Published when a PaxosAggregate is started.
        """

        __notifiable__ = False

    class AttributesChanged(Event):
        """
        Published when attributes of paxos_instance are changed.
        """

        __notifiable__ = False

        def __init__(self, changes: Optional[Dict[str, Any]] = None, **kwargs: Any):
            super(PaxosAggregate.AttributesChanged, self).__init__(
                changes=changes, **kwargs
            )

        @property
        def changes(self) -> Dict[str, Any]:
            return self.__dict__["changes"]

        def mutate(self, obj: "PaxosAggregate") -> None:
            for name, value in self.changes.items():
                setattr(obj, name, value)

    @classmethod
    def start(
        cls, originator_id: UUID, quorum_size: int, network_uid: str
    ) -> "PaxosAggregate":
        """
        Factory method that returns a new Paxos aggregate.
        """
        assert isinstance(quorum_size, int), "Not an integer: {}".format(quorum_size)
        started_class = cls.Started
        return cls.__create__(
            originator_id=originator_id,
            event_class=started_class,
            quorum_size=quorum_size,
            network_uid=network_uid,
        )

    def propose_value(self, value: Any, assume_leader: bool = False) -> PaxosMessage:
        """
        Proposes a value to the network.
        """
        if value is None:
            raise ValueError("Not allowed to propose value None")
        paxos = self.paxos_instance
        paxos.leader = assume_leader
        msg = paxos.propose_value(value)
        if msg is None:
            msg = paxos.prepare()
        self.setattrs_from_paxos(paxos)
        self.announce(msg)
        return msg

    def receive_message(self, msg: PaxosMessage) -> None:
        """
        Responds to messages from other participants.
        """
        if isinstance(msg, Resolution):
            return
        paxos = self.paxos_instance
        while msg:
            if isinstance(msg, Resolution):
                self.print_if_verbose(
                    "{} resolved value {}".format(self.network_uid, msg.value)
                )
                break
            else:
                self.print_if_verbose(
                    "{} <- {} <- {}".format(
                        self.network_uid, msg.__class__.__name__, msg.from_uid
                    )
                )
                msg = paxos.receive(msg)
                # Todo: Make it optional not to announce resolution
                #  (without which it's hard to see final value).
                do_announce_resolution = True
                if msg and (do_announce_resolution or not isinstance(msg, Resolution)):
                    self.announce(msg)

        self.setattrs_from_paxos(paxos)

    def announce(self, msg: PaxosMessage) -> None:
        """
        Announces a Paxos message.
        """
        self.print_if_verbose(
            "{} -> {}".format(self.network_uid, msg.__class__.__name__)
        )
        self.__trigger_event__(event_class=self.MessageAnnounced, msg=msg)

    class MessageAnnounced(Event):
        """
        Published when a Paxos message is announced.
        """

        @property
        def msg(self) -> PaxosMessage:
            return self.__dict__["msg"]

    def setattrs_from_paxos(self, paxos: PaxosInstance) -> None:
        """
        Registers changes of attribute value on Paxos instance.
        """
        changes = {}
        for name in self.paxos_variables:
            paxos_value = getattr(paxos, name)
            if paxos_value != getattr(self, name, None):
                self.print_if_verbose(
                    "{} {}: {}".format(self.network_uid, name, paxos_value)
                )
                changes[name] = paxos_value
                setattr(self, name, paxos_value)
        if changes:
            self.__trigger_event__(event_class=self.AttributesChanged, changes=changes)

    def print_if_verbose(self, param: Any) -> None:
        if self.is_verbose:
            print(param)

    def __str__(self) -> str:
        return (
            "PaxosAggregate("
            "final_value='{final_value}', "
            "proposed_value='{proposed_value}', "
            "network_uid='{network_uid}', "
            "proposal_id='{proposal_id}', "
            "promised_id='{promised_id}', "
            "promises_received='{promises_received}'"
            ")"
        ).format(**self.__dict__)


class PaxosApplication(ProcessApplication[PaxosAggregate, PaxosAggregate.Event]):
    persist_event_type = PaxosAggregate.Event
    quorum_size: int = 0
    notification_log_section_size = 5
    use_cache = True
    set_notification_ids = True

    @retry((RecordConflictError, OperationalError), max_attempts=10, wait=0)
    def propose_value(
        self, key: UUID, value: Any, assume_leader: bool = False
    ) -> PaxosAggregate:
        """
        Starts new Paxos aggregate and proposes a value for a key.

        Decorated with retry in case of notification log conflict
        or operational error.

        This a good example of writing process event from an
        application command. Just get the batch of pending events
        and record a process event with those events alone. They
        will be written atomically.
        """
        assert self.quorum_size > 0
        assert isinstance(key, UUID)
        paxos_aggregate = PaxosAggregate.start(
            originator_id=key, quorum_size=self.quorum_size, network_uid=self.name
        )
        msg = paxos_aggregate.propose_value(value, assume_leader=assume_leader)
        paxos_aggregate.receive_message(msg)

        self.save(paxos_aggregate)

        return paxos_aggregate  # in case it's new

    def get_final_value(self, key: UUID) -> PaxosAggregate:
        return self.repository[key].final_value

    def policy(
        self,
        repository: WrappedRepository[PaxosAggregate, PaxosAggregate.Event],
        event: PaxosAggregate.Event,
    ) -> Optional[Union[PaxosAggregate, Sequence[PaxosAggregate]]]:
        """
        Processes paxos "message announced" events of other applications
        by starting or continuing a paxos aggregate in this application.
        """
        if isinstance(event, PaxosAggregate.MessageAnnounced):
            # Get or create aggregate.
            try:
                paxos = repository[event.originator_id]
            except RepositoryKeyError:
                paxos = PaxosAggregate.start(
                    originator_id=event.originator_id,
                    quorum_size=self.quorum_size,
                    network_uid=self.name,
                )
                # Needs to go in the cache now, otherwise we get
                # "Duplicate" errors (for some unknown reason).
                if self.repository.use_cache:
                    self.repository.put_entity_in_cache(paxos.id, paxos)
            # Absolutely make sure the participant aggregates aren't getting confused.
            assert (
                paxos.network_uid == self.name
            ), "Wrong paxos aggregate: required network_uid {}, got {}".format(
                self.name, paxos.network_uid
            )
            # Only receive messages until resolution is
            # obtained. Followers will process our previous
            # announcements and resolve to the same final value
            # before processing anything we could announce after.
            msg = event.msg
            assert isinstance(msg, PaxosMessage)
            if paxos.final_value is None:
                paxos.receive_message(msg)

            return paxos
        else:
            return None


class PaxosSystem(System):
    def __init__(self, num_participants: int = 3, **kwargs: Any):
        self.num_participants = num_participants
        self.quorum_size = (num_participants + 2) // 2
        classes = [
            type(
                "PaxosApplication{}".format(i),
                (PaxosApplication,),
                {"quorum_size": self.quorum_size},
            )
            for i in range(num_participants)
        ]
        if num_participants > 1:
            pipelines = [[c[0], c[1], c[0]] for c in itertools.combinations(classes, 2)]
        else:
            pipelines = [classes]
        super(PaxosSystem, self).__init__(*pipelines, **kwargs)



# from eventsourcing.contrib.suffixtrees.domain.model.suffixtree import (
#     Edge,
#     EdgeRepository,
#     Node,
#     NodeRepository,
#     SuffixTree,
#     SuffixTreeRepository,
# )
# from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
#
#
# class SuffixTreeRepo(EventSourcedRepository, SuffixTreeRepository):
#     domain_class = SuffixTree
#
#
# class NodeRepo(EventSourcedRepository, NodeRepository):
#     domain_class = Node
#
#
# class EdgeRepo(EventSourcedRepository, EdgeRepository):
#     domain_class = Edge



# LONG_TEXT_CONT = """
# This article's use of external links may not follow Wikipedia's policies or guidelines. Please improve this
# article by removing excessive and inappropriate external links. (August 2010)
# Suffix Trees by Dr. Sartaj Sahni (CISE Department Chair at University of Florida)
# Suffix Trees by Lloyd Allison
# NIST's Dictionary of Algorithms and Data Structures: Suffix Tree
# suffix_tree ANSI C implementation of a Suffix Tree
# libstree, a generic suffix tree library written in C
# Tree::Suffix, a Perl binding to libstree
# Strmat a faster generic suffix tree library written in C (uses arrays instead of linked lists)
# SuffixTree a Python binding to Strmat
# Universal Data Compression Based on the Burrows-Wheeler Transformation: Theory and Practice, application of suffix
#  trees in the BWT
# Theory and Practice of Succinct Data Structures, C++ implementation of a compressed suffix tree]
# Practical Algorithm Template Library, a C++ library with suffix tree
#
# In computer science, a suffix tree (also called PAT tree or, in an earlier form, position tree) is a data
# structure that presents the suffixes of a given string in a way that allows for a particularly fast implementation
#  of many important string operations.
# The suffix tree for a string S is a tree whose edges are labeled with strings, such that each suffix of S
# corresponds to exactly one path from the tree's root to a leaf. It is thus a radix tree (more specifically,
# a Patricia tree) for the suffixes of S.
# Constructing such a tree for the string S takes time and space linear in the length of S. Once constructed,
# several operations can be performed quickly, for instance locating a substring in S, locating a substring if a
# certain number of mistakes are allowed, locating matches for a regular expression pattern etc. Suffix trees also
# provided one of the first linear-time solutions for the longest common substring problem. These speedups come at a
#  cost: storing a string's suffix tree typically requires significantly more space than storing the string itself.
# Contents [hide]
# 1 History
# 2 Definition
# 3 Generalised suffix tree
# 4 Functionality
# 5 Applications
# 6 Implementation
# 7 External construction
# 8 See also
# 9 References
# 10 External links
# [edit]History
#
# The concept was first introduced as a position tree by Weiner in 1973,[1] which Donald Knuth subsequently
# characterized as "Algorithm of the Year 1973". The construction was greatly simplified by McCreight in 1976 [2] ,
# and also by Ukkonen in 1995.[3][4] Ukkonen provided the first linear-time online-construction of suffix trees,
# now known as Ukkonen's algorithm. These algorithms are all linear-time for constant-size alphabet,
# and have worst-case running time of O(nlogn) in general.
# In 1997, Farach[5] gave the first suffix tree construction algorithm that is optimal for all alphabets. In
# particular, this is the first linear-time algorithm for strings drawn from an alphabet of integers in a polynomial
#  range. This latter algorithm has become the basis for new algorithms for constructing both suffix trees and
# suffix arrays, for example, in external memory, compressed, succinct, etc.
# [edit]Definition
#
# The suffix tree for the string S of length n is defined as a tree such that ([6] page 90):
# the paths from the root to the leaves have a one-to-one relationship with the suffixes of S,
# edges spell non-empty strings,
# and all internal nodes (except perhaps the root) have at least two children.
# Since such a tree does not exist for all strings, S is padded with a terminal symbol not seen in the string (
# usually denoted $). This ensures that no suffix is a prefix of another, and that there will be n leaf nodes,
# one for each of the n suffixes of S. Since all internal non-root nodes are branching, there can be at most n −  1
# such nodes, and n + (n − 1) + 1 = 2n nodes in total (n leaves, n − 1 internal nodes, 1 root).
# Suffix links are a key feature for older linear-time construction algorithms, although most newer algorithms,
# which are based on Farach's algorithm, dispense with suffix links. In a complete suffix tree, all internal
# non-root nodes have a suffix link to another internal node. If the path from the root to a node spells the string
# χα, where χ is a single character and α is a string (possibly empty), it has a suffix link to the internal node
# representing α. See for example the suffix link from the node for ANA to the node for NA in the figure above.
# Suffix links are also used in some algorithms running on the tree.
#
# A suffix tree for a string S of length n can be built in Θ(n) time, if the letters come from an alphabet of
# integers in a polynomial range (in particular, this is true for constant-sized alphabets).[5] For larger
# alphabets, the running time is dominated by first sorting the letters to bring them into a range of size O(n); in
# general, this takes O(nlogn) time. The costs below are given under the assumption that the alphabet is constant.
# Assume that a suffix tree has been built for the string S of length n, or that a generalised suffix tree has been
# built for the set of strings  of total length . You can:
# Search for strings:
# Check if a string P of length m is a substring in O(m) time ([6] page 92).
# Find the first occurrence of the patterns  of total length m as substrings in O(m) time.
# Find all z occurrences of the patterns  of total length m as substrings in O(m + z) time ([6] page 123).
# Search for a regular expression P in time expected sublinear in n ([7]).
# Find for each suffix of a pattern P, the length of the longest match between a prefix of  and a substring in D in
# Θ(m) time ([6] page 132). This is termed the matching statistics for P.
# Find properties of the strings:
# Find the longest common substrings of the string Si and Sj in Θ(ni + nj) time ([6] page 125).
# Find all maximal pairs, maximal repeats or supermaximal repeats in Θ(n + z) time ([6] page 144).
# Find the Lempel–Ziv decomposition in Θ(n) time ([6] page 166).
# Find the longest repeated substrings in Θ(n) time.
# Find the most frequently occurring substrings of a minimum length in Θ(n) time.
# Find the shortest strings from Σ that do not occur in D, in O(n + z) time, if there are z such strings.
# Find the shortest substrings occurring only once in Θ(n) time.
# Find, for each i, the shortest substrings of Si not occurring elsewhere in D in Θ(n) time.
# The suffix tree can be prepared for constant time lowest common ancestor retrieval between nodes in Θ(n) time ([6]
#  chapter 8). You can then also:
# Find the longest common prefix between the suffixes Si[p..ni] and Sj[q..nj] in Θ(1) ([6] page 196).
# Search for a pattern P of length m with at most k mismatches in O(kn + z) time, where z is the number of hits ([6]
#  page 200).
# Find all z maximal palindromes in Θ(n)([6] page 198), or Θ(gn) time if gaps of length g are allowed, or Θ(kn) if k
#  mismatches are allowed ([6] page 201).
# Find all z tandem repeats in O(nlogn + z), and k-mismatch tandem repeats in O(knlog(n / k) + z) ([6] page 204).
# Find the longest substrings common to at least k strings in D for  in Θ(n) time ([6] page 205).
# [edit]Applications
#
# Suffix trees can be used to solve a large number of string problems that occur in text-editing, free-text search,
# computational biology and other application areas.[8] Primary applications include:[8]
# String search, in O(m) complexity, where m is the length of the sub-string (but with initial O(n) time required to
#  build the suffix tree for the string)
# Finding the longest repeated substring
# Finding the longest common substring
# Finding the longest palindrome in a string
# Suffix trees are often used in bioinformatics applications, searching for patterns in DNA or protein sequences (
# which can be viewed as long strings of characters). The ability to search efficiently with mismatches might be
# considered their greatest strength. Suffix trees are also used in data compression; they can be used to find
# repeated data, and can be used for the sorting stage of the Burrows–Wheeler transform. Variants of the LZW
# compression schemes use suffix trees (LZSS). A suffix tree is also used in suffix tree clustering,
# a data clustering algorithm used in some search engines (first introduced in [9]).
# [edit]Implementation
#
# If each node and edge can be represented in Θ(1) space, the entire tree can be represented in Θ(n) space. The
# total length of all the strings on all of the edges in the tree is O(n2), but each edge can be stored as the
# position and length of a substring of S, giving a total space usage of Θ(n) computer words. The worst-case space
# usage of a suffix tree is seen with a fibonacci word, giving the full 2n nodes.
# An important choice when making a suffix tree implementation is the parent-child relationships between nodes. The
# most common is using linked lists called sibling lists. Each node has a pointer to its first child, and to the
# next node in the child list it is a part of. Hash maps, sorted/unsorted arrays (with array doubling), and balanced
#  search trees may also be used, giving different running time properties. We are interested in:
# The cost of finding the child on a given character.
# The cost of inserting a child.
# The cost of enlisting all children of a node (divided by the number of children in the table below).
# Let σ be the size of the alphabet. Then you have the following costs:
# Lookup	Insertion	Traversal
# Sibling lists / unsorted arrays	O(σ)	Θ(1)	Θ(1)
# Hash maps	Θ(1)	Θ(1)	O(σ)
# Balanced search tree	O(logσ)	O(logσ)	O(1)
# Sorted arrays	O(logσ)	O(σ)	O(1)
# Hash maps + sibling lists	O(1)	O(1)	O(1)
# Note that the insertion cost is amortised, and that the costs for hashing are given perfect hashing.
# The large amount of information in each edge and node makes the suffix tree very expensive, consuming about ten to
#  twenty times the memory size of the source text in good implementations. The suffix array reduces this
# requirement to a factor of four, and researchers have continued to find smaller indexing structures.
# [edit]External construction
#
# Suffix trees quickly outgrow the main memory on standard machines for sequence collections in the order of
# gigabytes. As such, their construction calls for external memory approaches.
# There are theoretical results for constructing suffix trees in external memory. The algorithm by Farach et al. [
# 10] is theoretically optimal, with an I/O complexity equal to that of sorting. However, as discussed for example
# in ,[11] the overall intricacy of this algorithm has prevented, so far, its practical implementation.
# On the other hand, there have been practical works for constructing disk-based suffix trees which scale to (few)
# GB/hours. The state of the art methods are TDD ,[12] TRELLIS [13] , DiGeST ,[14] and B2ST .[15]
# TDD and TRELLIS scale up to the entire human genome – approximately 3GB – resulting in a disk-based suffix tree of
#  a size in the tens of gigabytes,.[12][13] However, these methods cannot handle efficiently collections of
# sequences exceeding 3GB.[14] DiGeST performs significantly better and is able to handle collections of sequences
# in the order of 6GB in about 6 hours.[14] The source code and documentation for the latter is available from [16]
# . All these methods can efficiently build suffix trees for the case when the tree does not fit in main memory,
# but the input does. The most recent method, B2ST,[15] scales to handle inputs that do not fit in main memory.
# [edit]See also
#
# Suffix array
# Generalised suffix tree
# [edit]References
#
# ^ P. Weiner (1973). "Linear pattern matching algorithm". 14th Annual IEEE Symposium on Switching and Automata
# Theory. pp. 1–11. doi:10.1109/SWAT.1973.13.
# ^ Edward M. McCreight (1976). "A Space-Economical Suffix Tree Construction Algorithm". Journal of the ACM 23 (2):
# 262–272. doi:10.1145/321941.321946.
# ^ E. Ukkonen (1995). "On-line construction of suffix trees". Algorithmica 14 (3): 249–260. doi:10.1007/BF01206331.
# ^ R. Giegerich and S. Kurtz (1997). "From Ukkonen to McCreight and Weiner: A Unifying View of Linear-Time Suffix
# Tree Construction". Algorithmica 19 (3): 331–353. doi:10.1007/PL00009177.
# ^ a b M. Farach (1997). "Optimal Suffix Tree Construction with Large Alphabets". FOCS: 137–143.
# ^ a b c d e f g h i j k l m n Gusfield, Dan (1999) [1997]. Algorithms on Strings, Trees and Sequences: Computer
# Science and Computational Biology. USA: Cambridge University Press. ISBN 0-521-58519-8.
# ^ Ricardo A. Baeza-Yates and Gaston H. Gonnet (1996). "Fast text searching for regular expressions or automaton
# searching on tries". Journal of the ACM (ACM Press) 43 (6): 915–936. doi:10.1145/235809.235810.
# ^ a b Allison, L.. "Suffix Trees". Retrieved 2008-10-14.
# ^ Oren Zamir and Oren Etzioni (1998). "Web document clustering: a feasibility demonstration". SIGIR '98:
# Proceedings of the 21st annual international ACM SIGIR conference on Research and development in information
# retrieval. ACM. pp. 46–54.
# ^ Martin Farach-Colton, Paolo Ferragina, S. Muthukrishnan (2000). "On the sorting-complexity of suffix tree
# construction.". J. Acm 47(6) 47 (6): 987–1011. doi:10.1145/355541.355547.
# ^ Smyth, William (2003). Computing Patterns in Strings. Addison-Wesley.
# ^ a b Sandeep Tata, Richard A. Hankins, and Jignesh M. Patel (2003). "Practical Suffix Tree Construction". VLDB
# '03: Proceedings of the 30th International Conference on Very Large Data Bases. Morgan Kaufmann. pp. 36–47.
# ^ a b Benjarath Phoophakdee and Mohammed J. Zaki (2007). "Genome-scale disk-based suffix tree indexing". SIGMOD
# '07: Proceedings of the ACM SIGMOD International Conference on Management of Data. ACM. pp. 833–844.
# ^ a b c Marina Barsky, Ulrike Stege, Alex Thomo, and Chris Upton (2008). "A new method for indexing genomes using
# on-disk suffix trees". CIKM '08: Proceedings of the 17th ACM Conference on Information and Knowledge Management.
# ACM. pp. 649–658.
# ^ a b Marina Barsky, Ulrike Stege, Alex Thomo, and Chris Upton (2009). "Suffix trees for very large genomic
# sequences". CIKM '09: Proceedings of the 18th ACM Conference on Information and Knowledge Management. ACM.
# ^ "The disk-based suffix tree for pattern search in sequenced genomes". Retrieved 2009-10-15.
# [edit]External links
#
#
# This article's use of external links may not follow Wikipedia's policies or guidelines. Please improve this
# article by removing excessive and inappropriate external links. (August 2010)
# Suffix Trees by Dr. Sartaj Sahni (CISE Department Chair at University of Florida)
# Suffix Trees by Lloyd Allison
# NIST's Dictionary of Algorithms and Data Structures: Suffix Tree
# suffix_tree ANSI C implementation of a Suffix Tree
# libstree, a generic suffix tree library written in C
# Tree::Suffix, a Perl binding to libstree
# Strmat a faster generic suffix tree library written in C (uses arrays instead of linked lists)
# SuffixTree a Python binding to Strmat
# Universal Data Compression Based on the Burrows-Wheeler Transformation: Theory and Practice, application of suffix
#  trees in the BWT
# Theory and Practice of Succinct Data Structures, C++ implementation of a compressed suffix tree]
# Practical Algorithm Template Library, a C++ library with suffix treeIn computer science, a suffix tree (also
# called PAT tree or, in an earlier form, position tree) is a data structure that presents the suffixes of a given
# string in a way that allows for a particularly fast implementation of many important string operations.
# The suffix tree for a string S is a tree whose edges are labeled with strings, such that each suffix of S
# corresponds to exactly one path from the tree's root to a leaf. It is thus a radix tree (more specifically,
# a Patricia tree) for the suffixes of S.
# Constructing such a tree for the string S takes time and space linear in the length of S. Once constructed,
# several operations can be performed quickly, for instance locating a substring in S, locating a substring if a
# certain number of mistakes are allowed, locating matches for a regular expression pattern etc. Suffix trees also
# provided one of the first linear-time solutions for the longest common substring problem. These speedups come at a
#  cost: storing a string's suffix tree typically requires significantly more space than storing the string itself.
# Contents [hide]
# 1 History
# 2 Definition
# 3 Generalised suffix tree
# 4 Functionality
# 5 Applications
# 6 Implementation
# 7 External construction
# 8 See also
# 9 References
# 10 External links
# [edit]History
#
# The concept was first introduced as a position tree by Weiner in 1973,[1] which Donald Knuth subsequently
# characterized as "Algorithm of the Year 1973". The construction was greatly simplified by McCreight in 1976 [2] ,
# and also by Ukkonen in 1995.[3][4] Ukkonen provided the first linear-time online-construction of suffix trees,
# now known as Ukkonen's algorithm. These algorithms are all linear-time for constant-size alphabet,
# and have worst-case running time of O(nlogn) in general.
# In 1997, Farach[5] gave the first suffix tree construction algorithm that is optimal for all alphabets. In
# particular, this is the first linear-time algorithm for strings drawn from an alphabet of integers in a polynomial
#  range. This latter algorithm has become the basis for new algorithms for constructing both suffix trees and
# suffix arrays, for example, in external memory, compressed, succinct, etc.
# [edit]Definition
#
# The suffix tree for the string S of length n is defined as a tree such that ([6] page 90):
# the paths from the root to the leaves have a one-to-one relationship with the suffixes of S,
# edges spell non-empty strings,
# and all internal nodes (except perhaps the root) have at least two children.
# Since such a tree does not exist for all strings, S is padded with a terminal symbol not seen in the string (
# usually denoted $). This ensures that no suffix is a prefix of another, and that there will be n leaf nodes,
# one for each of the n suffixes of S. Since all internal non-root nodes are branching, there can be at most n −  1
# such nodes, and n + (n − 1) + 1 = 2n nodes in total (n leaves, n − 1 internal nodes, 1 root).
# Suffix links are a key feature for older linear-time construction algorithms, although most newer algorithms,
# which are based on Farach's algorithm, dispense with suffix links. In a complete suffix tree, all internal
# non-root nodes have a suffix link to another internal node. If the path from the root to a node spells the string
# χα, where χ is a single character and α is a string (possibly empty), it has a suffix link to the internal node
# representing α. See for example the suffix link from the node for ANA to the node for NA in the figure above.
# Suffix links are also used in some algorithms running on the tree.
#
# A suffix tree for a string S of length n can be built in Θ(n) time, if the letters come from an alphabet of
# integers in a polynomial range (in particular, this is true for constant-sized alphabets).[5] For larger
# alphabets, the running time is dominated by first sorting the letters to bring them into a range of size O(n); in
# general, this takes O(nlogn) time. The costs below are given under the assumption that the alphabet is constant.
# Assume that a suffix tree has been built for the string S of length n, or that a generalised suffix tree has been
# built for the set of strings  of total length . You can:
# Search for strings:
# Check if a string P of length m is a substring in O(m) time ([6] page 92).
# Find the first occurrence of the patterns  of total length m as substrings in O(m) time.
# Find all z occurrences of the patterns  of total length m as substrings in O(m + z) time ([6] page 123).
# Search for a regular expression P in time expected sublinear in n ([7]).
# Find for each suffix of a pattern P, the length of the longest match between a prefix of  and a substring in D in
# Θ(m) time ([6] page 132). This is termed the matching statistics for P.
# Find properties of the strings:
# Find the longest common substrings of the string Si and Sj in Θ(ni + nj) time ([6] page 125).
# Find all maximal pairs, maximal repeats or supermaximal repeats in Θ(n + z) time ([6] page 144).
# Find the Lempel–Ziv decomposition in Θ(n) time ([6] page 166).
# Find the longest repeated substrings in Θ(n) time.
# Find the most frequently occurring substrings of a minimum length in Θ(n) time.
# Find the shortest strings from Σ that do not occur in D, in O(n + z) time, if there are z such strings.
# Find the shortest substrings occurring only once in Θ(n) time.
# Find, for each i, the shortest substrings of Si not occurring elsewhere in D in Θ(n) time.
# The suffix tree can be prepared for constant time lowest common ancestor retrieval between nodes in Θ(n) time ([6]
#  chapter 8). You can then also:
# Find the longest common prefix between the suffixes Si[p..ni] and Sj[q..nj] in Θ(1) ([6] page 196).
# Search for a pattern P of length m with at most k mismatches in O(kn + z) time, where z is the number of hits ([6]
#  page 200).
# Find all z maximal palindromes in Θ(n)([6] page 198), or Θ(gn) time if gaps of length g are allowed, or Θ(kn) if k
#  mismatches are allowed ([6] page 201).
# Find all z tandem repeats in O(nlogn + z), and k-mismatch tandem repeats in O(knlog(n / k) + z) ([6] page 204).
# Find the longest substrings common to at least k strings in D for  in Θ(n) time ([6] page 205).
# [edit]Applications
#
# Suffix trees can be used to solve a large number of string problems that occur in text-editing, free-text search,
# computational biology and other application areas.[8] Primary applications include:[8]
# String search, in O(m) complexity, where m is the length of the sub-string (but with initial O(n) time required to
#  build the suffix tree for the string)
# Finding the longest repeated substring
# Finding the longest common substring
# Finding the longest palindrome in a string
# Suffix trees are often used in bioinformatics applications, searching for patterns in DNA or protein sequences (
# which can be viewed as long strings of characters). The ability to search efficiently with mismatches might be
# considered their greatest strength. Suffix trees are also used in data compression; they can be used to find
# repeated data, and can be used for the sorting stage of the Burrows–Wheeler transform. Variants of the LZW
# compression schemes use suffix trees (LZSS). A suffix tree is also used in suffix tree clustering,
# a data clustering algorithm used in some search engines (first introduced in [9]).
# [edit]Implementation
#
# If each node and edge can be represented in Θ(1) space, the entire tree can be represented in Θ(n) space. The
# total length of all the strings on all of the edges in the tree is O(n2), but each edge can be stored as the
# position and length of a substring of S, giving a total space usage of Θ(n) computer words. The worst-case space
# usage of a suffix tree is seen with a fibonacci word, giving the full 2n nodes.
# An important choice when making a suffix tree implementation is the parent-child relationships between nodes. The
# most common is using linked lists called sibling lists. Each node has a pointer to its first child, and to the
# next node in the child list it is a part of. Hash maps, sorted/unsorted arrays (with array doubling), and balanced
#  search trees may also be used, giving different running time properties. We are interested in:
# The cost of finding the child on a given character.
# The cost of inserting a child.
# The cost of enlisting all children of a node (divided by the number of children in the table below).
# Let σ be the size of the alphabet. Then you have the following costs:
# Lookup	Insertion	Traversal
# Sibling lists / unsorted arrays	O(σ)	Θ(1)	Θ(1)
# Hash maps	Θ(1)	Θ(1)	O(σ)
# Balanced search tree	O(logσ)	O(logσ)	O(1)
# Sorted arrays	O(logσ)	O(σ)	O(1)
# Hash maps + sibling lists	O(1)	O(1)	O(1)
# Note that the insertion cost is amortised, and that the costs for hashing are given perfect hashing.
# The large amount of information in each edge and node makes the suffix tree very expensive, consuming about ten to
#  twenty times the memory size of the source text in good implementations. The suffix array reduces this
# requirement to a factor of four, and researchers have continued to find smaller indexing structures.
# [edit]External construction
#
# Suffix trees quickly outgrow the main memory on standard machines for sequence collections in the order of
# gigabytes. As such, their construction calls for external memory approaches.
# There are theoretical results for constructing suffix trees in external memory. The algorithm by Farach et al. [
# 10] is theoretically optimal, with an I/O complexity equal to that of sorting. However, as discussed for example
# in ,[11] the overall intricacy of this algorithm has prevented, so far, its practical implementation.
# On the other hand, there have been practical works for constructing disk-based suffix trees which scale to (few)
# GB/hours. The state of the art methods are TDD ,[12] TRELLIS [13] , DiGeST ,[14] and B2ST .[15]
# TDD and TRELLIS scale up to the entire human genome – approximately 3GB – resulting in a disk-based suffix tree of
#  a size in the tens of gigabytes,.[12][13] However, these methods cannot handle efficiently collections of
# sequences exceeding 3GB.[14] DiGeST performs significantly better and is able to handle collections of sequences
# in the order of 6GB in about 6 hours.[14] The source code and documentation for the latter is available from [16]
# . All these methods can efficiently build suffix trees for the case when the tree does not fit in main memory,
# but the input does. The most recent method, B2ST,[15] scales to handle inputs that do not fit in main memory.
# [edit]See also
#
# Suffix array
# Generalised suffix tree
# [edit]References
#
# ^ P. Weiner (1973). "Linear pattern matching algorithm". 14th Annual IEEE Symposium on Switching and Automata
# Theory. pp. 1–11. doi:10.1109/SWAT.1973.13.
# ^ Edward M. McCreight (1976). "A Space-Economical Suffix Tree Construction Algorithm". Journal of the ACM 23 (2):
# 262–272. doi:10.1145/321941.321946.
# ^ E. Ukkonen (1995). "On-line construction of suffix trees". Algorithmica 14 (3): 249–260. doi:10.1007/BF01206331.
# ^ R. Giegerich and S. Kurtz (1997). "From Ukkonen to McCreight and Weiner: A Unifying View of Linear-Time Suffix
# Tree Construction". Algorithmica 19 (3): 331–353. doi:10.1007/PL00009177.
# ^ a b M. Farach (1997). "Optimal Suffix Tree Construction with Large Alphabets". FOCS: 137–143.
# ^ a b c d e f g h i j k l m n Gusfield, Dan (1999) [1997]. Algorithms on Strings, Trees and Sequences: Computer
# Science and Computational Biology. USA: Cambridge University Press. ISBN 0-521-58519-8.
# ^ Ricardo A. Baeza-Yates and Gaston H. Gonnet (1996). "Fast text searching for regular expressions or automaton
# searching on tries". Journal of the ACM (ACM Press) 43 (6): 915–936. doi:10.1145/235809.235810.
# ^ a b Allison, L.. "Suffix Trees". Retrieved 2008-10-14.
# ^ Oren Zamir and Oren Etzioni (1998). "Web document clustering: a feasibility demonstration". SIGIR '98:
# Proceedings of the 21st annual international ACM SIGIR conference on Research and development in information
# retrieval. ACM. pp. 46–54.
# ^ Martin Farach-Colton, Paolo Ferragina, S. Muthukrishnan (2000). "On the sorting-complexity of suffix tree
# construction.". J. Acm 47(6) 47 (6): 987–1011. doi:10.1145/355541.355547.
# ^ Smyth, William (2003). Computing Patterns in Strings. Addison-Wesley.
# ^ a b Sandeep Tata, Richard A. Hankins, and Jignesh M. Patel (2003). "Practical Suffix Tree Construction". VLDB
# '03: Proceedings of the 30th International Conference on Very Large Data Bases. Morgan Kaufmann. pp. 36–47.
# ^ a b Benjarath Phoophakdee and Mohammed J. Zaki (2007). "Genome-scale disk-based suffix tree indexing". SIGMOD
# '07: Proceedings of the ACM SIGMOD International Conference on Management of Data. ACM. pp. 833–844.
# ^ a b c Marina Barsky, Ulrike Stege, Alex Thomo, and Chris Upton (2008). "A new method for indexing genomes using
# on-disk suffix trees". CIKM '08: Proceedings of the 17th ACM Conference on Information and Knowledge Management.
# ACM. pp. 649–658.
# ^ a b Marina Barsky, Ulrike Stege, Alex Thomo, and Chris Upton (2009). "Suffix trees for very large genomic
# sequences". CIKM '09: Proceedings of the 18th ACM Conference on Information and Knowledge Management. ACM.
# ^ "The disk-based suffix tree for pattern search in sequenced genomes". Retrieved 2009-10-15.
# [edit]External links
# """
# LONG_TEXT = """In computer science, a suffix tree (also called PAT tree or, in an earlier form, position tree) is
# a data structure that presents the suffixes of a given string in a way that allows for a particularly fast
# implementation of many important string operations.
# The suffix tree for a string S is a tree whose edges are labeled with strings, such that each suffix of S
# corresponds to exactly one path from the tree's root to a leaf. It is thus a radix tree (more specifically,
# a Patricia tree) for the suffixes of S.
# Constructing such a tree for the string S takes time and space linear in the length of S. Once constructed,
# several operations can be performed quickly, for instance locating a substring in S, locating a substring if a
# certain number of mistakes are allowed, locating matches for a regular expression pattern etc. Suffix trees also
# provided one of the first linear-time solutions for the longest common substring problem. These speedups come at a
#  cost: storing a string's suffix tree typically requires significantly more space than storing the string itself.
# Contents [hide]
# 1 History
# 2 Definition
# 3 Generalised suffix tree
# 4 Functionality
# 5 Applications
# 6 Implementation
# 7 External construction
# 8 See also
# 9 References
# 10 External links
# [edit]History
#
# The concept was first introduced as a position tree by Weiner in 1973,[1] which Donald Knuth subsequently
# characterized as "Algorithm of the Year 1973". The construction was greatly simplified by McCreight in 1976 [2] ,
# and also by Ukkonen in 1995.[3][4] Ukkonen provided the first linear-time online-construction of suffix trees,
# now known as Ukkonen's algorithm. These algorithms are all linear-time for constant-size alphabet,
# and have worst-case running time of O(nlogn) in general.
# In 1997, Farach[5] gave the first suffix tree construction algorithm that is optimal for all alphabets. In
# particular, this is the first linear-time algorithm for strings drawn from an alphabet of integers in a polynomial
#  range. This latter algorithm has become the basis for new algorithms for constructing both suffix trees and
# suffix arrays, for example, in external memory, compressed, succinct, etc.
# [edit]Definition
#
# The suffix tree for the string S of length n is defined as a tree such that ([6] page 90):
# the paths from the root to the leaves have a one-to-one relationship with the suffixes of S,
# edges spell non-empty strings,
# and all internal nodes (except perhaps the root) have at least two children.
# Since such a tree does not exist for all strings, S is padded with a terminal symbol not seen in the string (
# usually denoted $). This ensures that no suffix is a prefix of another, and that there will be n leaf nodes,
# one for each of the n suffixes of S. Since all internal non-root nodes are branching, there can be at most n −  1
# such nodes, and n + (n − 1) + 1 = 2n nodes in total (n leaves, n − 1 internal nodes, 1 root).
# Suffix links are a key feature for older linear-time construction algorithms, although most newer algorithms,
# which are based on Farach's algorithm, dispense with suffix links. In a complete suffix tree, all internal
# non-root nodes have a suffix link to another internal node. If the path from the root to a node spells the string
# χα, where χ is a single character and α is a string (possibly empty), it has a suffix link to the internal node
# representing α. See for example the suffix link from the node for ANA to the node for NA in the figure above.
# Suffix links are also used in some algorithms running on the tree.
#
# A suffix tree for a string S of length n can be built in Θ(n) time, if the letters come from an alphabet of
# integers in a polynomial range (in particular, this is true for constant-sized alphabets).[5] For larger
# alphabets, the running time is dominated by first sorting the letters to bring them into a range of size O(n); in
# general, this takes O(nlogn) time. The costs below are given under the assumption that the alphabet is constant.
# Assume that a suffix tree has been built for the string S of length n, or that a generalised suffix tree has been
# built for the set of strings  of total length . You can:
# Search for strings:
# Check if a string P of length m is a substring in O(m) time ([6] page 92).
# Find the first occurrence of the patterns  of total length m as substrings in O(m) time.
# Find all z occurrences of the patterns  of total length m as substrings in O(m + z) time ([6] page 123).
# Search for a regular expression P in time expected sublinear in n ([7]).
# Find for each suffix of a pattern P, the length of the longest match between a prefix of  and a substring in D in
# Θ(m) time ([6] page 132). This is termed the matching statistics for P.
# Find properties of the strings:
# Find the longest common substrings of the string Si and Sj in Θ(ni + nj) time ([6] page 125).
# Find all maximal pairs, maximal repeats or supermaximal repeats in Θ(n + z) time ([6] page 144).
# Find the Lempel–Ziv decomposition in Θ(n) time ([6] page 166).
# Find the longest repeated substrings in Θ(n) time.
# Find the most frequently occurring substrings of a minimum length in Θ(n) time.
# Find the shortest strings from Σ that do not occur in D, in O(n + z) time, if there are z such strings.
# Find the shortest substrings occurring only once in Θ(n) time.
# Find, for each i, the shortest substrings of Si not occurring elsewhere in D in Θ(n) time.
# The suffix tree can be prepared for constant time lowest common ancestor retrieval between nodes in Θ(n) time ([6]
#  chapter 8). You can then also:
# Find the longest common prefix between the suffixes Si[p..ni] and Sj[q..nj] in Θ(1) ([6] page 196).
# Search for a pattern P of length m with at most k mismatches in O(kn + z) time, where z is the number of hits ([6]
#  page 200).
# Find all z maximal palindromes in Θ(n)([6] page 198), or Θ(gn) time if gaps of length g are allowed, or Θ(kn) if k
#  mismatches are allowed ([6] page 201).
# Find all z tandem repeats in O(nlogn + z), and k-mismatch tandem repeats in O(knlog(n / k) + z) ([6] page 204).
# Find the longest substrings common to at least k strings in D for  in Θ(n) time ([6] page 205).
# [edit]Applications
#
# Suffix trees can be used to solve a large number of string problems that occur in text-editing, free-text search,
# computational biology and other application areas.[8] Primary applications include:[8]
# String search, in O(m) complexity, where m is the length of the sub-string (but with initial O(n) time required to
#  build the suffix tree for the string)
# Finding the longest repeated substring
# Finding the longest common substring
# Finding the longest palindrome in a string
# Suffix trees are often used in bioinformatics applications, searching for patterns in DNA or protein sequences (
# which can be viewed as long strings of characters). The ability to search efficiently with mismatches might be
# considered their greatest strength. Suffix trees are also used in data compression; they can be used to find
# repeated data, and can be used for the sorting stage of the Burrows–Wheeler transform. Variants of the LZW
# compression schemes use suffix trees (LZSS). A suffix tree is also used in suffix tree clustering,
# a data clustering algorithm used in some search engines (first introduced in [9]).
# [edit]Implementation
#
# If each node and edge can be represented in Θ(1) space, the entire tree can be represented in Θ(n) space. The
# total length of all the strings on all of the edges in the tree is O(n2), but each edge can be stored as the
# position and length of a substring of S, giving a total space usage of Θ(n) computer words. The worst-case space
# usage of a suffix tree is seen with a fibonacci word, giving the full 2n nodes.
# An important choice when making a suffix tree implementation is the parent-child relationships between nodes. The
# most common is using linked lists called sibling lists. Each node has a pointer to its first child, and to the
# next node in the child list it is a part of. Hash maps, sorted/unsorted arrays (with array doubling), and balanced
#  search trees may also be used, giving different running time properties. We are interested in:
# The cost of finding the child on a given character.
# The cost of inserting a child.
# The cost of enlisting all children of a node (divided by the number of children in the table below).
# Let σ be the size of the alphabet. Then you have the following costs:
# Lookup	Insertion	Traversal
# Sibling lists / unsorted arrays	O(σ)	Θ(1)	Θ(1)
# Hash maps	Θ(1)	Θ(1)	O(σ)
# Balanced search tree	O(logσ)	O(logσ)	O(1)
# Sorted arrays	O(logσ)	O(σ)	O(1)
# Hash maps + sibling lists	O(1)	O(1)	O(1)
# Note that the insertion cost is amortised, and that the costs for hashing are given perfect hashing.
# The large amount of information in each edge and node makes the suffix tree very expensive, consuming about ten to
#  twenty times the memory size of the source text in good implementations. The suffix array reduces this
# requirement to a factor of four, and researchers have continued to find smaller indexing structures.
# [edit]External construction
#
# Suffix trees quickly outgrow the main memory on standard machines for sequence collections in the order of
# gigabytes. As such, their construction calls for external memory approaches.
# There are theoretical results for constructing suffix trees in external memory. The algorithm by Farach et al. [
# 10] is theoretically optimal, with an I/O complexity equal to that of sorting. However, as discussed for example
# in ,[11] the overall intricacy of this algorithm has prevented, so far, its practical implementation.
# On the other hand, there have been practical works for constructing disk-based suffix trees which scale to (few)
# GB/hours. The state of the art methods are TDD ,[12] TRELLIS [13] , DiGeST ,[14] and B2ST .[15]
# TDD and TRELLIS scale up to the entire human genome – approximately 3GB – resulting in a disk-based suffix tree of
#  a size in the tens of gigabytes,.[12][13] However, these methods cannot handle efficiently collections of
# sequences exceeding 3GB.[14] DiGeST performs significantly better and is able to handle collections of sequences
# in the order of 6GB in about 6 hours.[14] The source code and documentation for the latter is available from [16]
# . All these methods can efficiently build suffix trees for the case when the tree does not fit in main memory,
# but the input does. The most recent method, B2ST,[15] scales to handle inputs that do not fit in main memory.
# [edit]See also
#
# Suffix array
# Generalised suffix tree
# [edit]References
#
# ^ P. Weiner (1973). "Linear pattern matching algorithm". 14th Annual IEEE Symposium on Switching and Automata
# Theory. pp. 1–11. doi:10.1109/SWAT.1973.13.
# ^ Edward M. McCreight (1976). "A Space-Economical Suffix Tree Construction Algorithm". Journal of the ACM 23 (2):
# 262–272. doi:10.1145/321941.321946.
# ^ E. Ukkonen (1995). "On-line construction of suffix trees". Algorithmica 14 (3): 249–260. doi:10.1007/BF01206331.
# ^ R. Giegerich and S. Kurtz (1997). "From Ukkonen to McCreight and Weiner: A Unifying View of Linear-Time Suffix
# Tree Construction". Algorithmica 19 (3): 331–353. doi:10.1007/PL00009177.
# ^ a b M. Farach (1997). "Optimal Suffix Tree Construction with Large Alphabets". FOCS: 137–143.
# ^ a b c d e f g h i j k l m n Gusfield, Dan (1999) [1997]. Algorithms on Strings, Trees and Sequences: Computer
# Science and Computational Biology. USA: Cambridge University Press. ISBN 0-521-58519-8.
# ^ Ricardo A. Baeza-Yates and Gaston H. Gonnet (1996). "Fast text searching for regular expressions or automaton
# searching on tries". Journal of the ACM (ACM Press) 43 (6): 915–936. doi:10.1145/235809.235810.
# ^ a b Allison, L.. "Suffix Trees". Retrieved 2008-10-14.
# ^ Oren Zamir and Oren Etzioni (1998). "Web document clustering: a feasibility demonstration". SIGIR '98:
# Proceedings of the 21st annual international ACM SIGIR conference on Research and development in information
# retrieval. ACM. pp. 46–54.
# ^ Martin Farach-Colton, Paolo Ferragina, S. Muthukrishnan (2000). "On the sorting-complexity of suffix tree
# construction.". J. Acm 47(6) 47 (6): 987–1011. doi:10.1145/355541.355547.
# ^ Smyth, William (2003). Computing Patterns in Strings. Addison-Wesley.
# ^ a b Sandeep Tata, Richard A. Hankins, and Jignesh M. Patel (2003). "Practical Suffix Tree Construction". VLDB
# '03: Proceedings of the 30th International Conference on Very Large Data Bases. Morgan Kaufmann. pp. 36–47.
# ^ a b Benjarath Phoophakdee and Mohammed J. Zaki (2007). "Genome-scale disk-based suffix tree indexing". SIGMOD
# '07: Proceedings of the ACM SIGMOD International Conference on Management of Data. ACM. pp. 833–844.
# ^ a b c Marina Barsky, Ulrike Stege, Alex Thomo, and Chris Upton (2008). "A new method for indexing genomes using
# on-disk suffix trees". CIKM '08: Proceedings of the 17th ACM Conference on Information and Knowledge Management.
# ACM. pp. 649–658.
# ^ a b Marina Barsky, Ulrike Stege, Alex Thomo, and Chris Upton (2009). "Suffix trees for very large genomic
# sequences". CIKM '09: Proceedings of the 18th ACM Conference on Information and Knowledge Management. ACM.
# ^ "The disk-based suffix tree for pattern search in sequenced genomes". Retrieved 2009-10-15.
# [edit]External links
# """
